# IPython log file

# import swipe
import numpy as np

from pyswipe.swipe import SWIPE

plotdir = '/plotdir/'

import matplotlib as mpl
mplBkgrnd = 'QtAgg'
mpl.use(mplBkgrnd)

# MPL opts

mpl.rcParams.update({'text.color': 'k'})
mpl.rcParams.update({'axes.labelcolor': 'k'})
mpl.rcParams.update({'xtick.color': 'k'})
mpl.rcParams.update({'ytick.color': 'k'})
mpl.rcParams.update({'text.usetex': False})

mpl.rcParams.update({'figure.figsize': [10.0, 8.0]})
mpl.rcParams.update({'savefig.directory': plotdir})

import matplotlib.pyplot as plt
plt.ion()

from polplot.polplot import Polarplot

d2r = np.pi/180

no_SH = True
##############################

B = 5                          # nT

ca = {(0, 1): 0, (0, 0):-45, (1, 0):-90,(2,0):-135,(2,1):180,(0,2):45,(1,2):90,(2,2):135}

tilts = [-25, 0, 25]
tilts = [0]

v = 450
f107val = 120

minlat = 47
minlat_for_cpcp_calc = 60

# potlevels = np.linspace(-.95, .95, 20)/2
potlevels = np.arange(-55, 55, 5)+2.5


doCalcBhattacharyya = False      # Code for this taken from "pct_diff_ns_currents_for_varying_vswB_and_clock_angle.py"

##############################
# Initialize models
swipe_n = SWIPE(v, B, B, 0, f107val,
             minlat=minlat)# ,
             # use_transpose_coeff_fn=False,
             # coeff_fn=coeff_fn,
             # zero_lats=zero_latNU)
swipe_s = SWIPE(v, B, B, 0, f107val,
             minlat=minlat)# ,
             # use_transpose_coeff_fn=False,
             # coeff_fn=coeff_fn,
             # zero_lats=zero_latNU)

mlat_n, mlt_n = map(lambda x:np.split(x, 2)[0], swipe_n.scalargrid)
mlat_s, mlt_s = map(lambda x:np.split(x, 2)[1], swipe_n.scalargrid)

OKindsn = np.abs(mlat_n) >= minlat_for_cpcp_calc
OKindss = np.abs(mlat_s) >= minlat_for_cpcp_calc

dPhidn = dict()
dPhids = dict()
locdn = dict()
locds = dict()
    

if doCalcBhattacharyya:
    # Ripped from pyamps.amps.get_integrated_upward_current
    REFRE = 6371.2 # Reference radius used in geomagnetic modeling

    # get surface area element in each cell:
    mlat, mlt = swipe_n.scalargrid
    mlt_sorted = np.sort(np.unique(mlt))
    mltres = (mlt_sorted[1] - mlt_sorted[0]) * np.pi / 12
    mlat_sorted = np.sort(np.unique(mlat))
    mlatres = (mlat_sorted[1] - mlat_sorted[0]) * np.pi / 180
    R = (REFRE + swipe_n.height) * 1e3  # radius in meters
    dS = R**2 * np.cos(mlat * np.pi/180) * mlatres * mltres

    dS_n,dS_s = np.split(dS, 2)
    assert np.array_equal(dS_n,dS_s),"Uh oh, calculations later assume dS_n == dS_s!"
    dS = dS_n

    def sep_into_pos_and_neg_dist_and_norm(vals,dS):
        valspos, valsneg = np.copy(vals),-vals
        valspos[valspos < 0] = 0
        valsneg[valsneg < 0] = 0
        
        vals_sum = np.sum(valspos*dS+valsneg*dS)
        
        valspos = valspos/vals_sum
        valsneg = valsneg/vals_sum
        
        return valspos,valsneg


##############################
# Loop
# lists for holding CPCP for each bz value
cpcpn, cpcps = [], []
locns, locss = [], []
fig = plt.figure(1,figsize = (9, 9))

for tilt in tilts:#tilt = tilts[0]

    plt.clf()

    axes = [[Polarplot(plt.subplot2grid((91, 3), (j*30, i), rowspan = 30),
                          linestyle = ':', color = 'grey', minlat = minlat, linewidth = .9)
             for i in range(3) if (i, j) != (1, 1)]
             for j in range(3)]
    axdial = plt.subplot2grid((91, 3), (30, 1), rowspan = 30)
    axdial.set_aspect('equal')
    axdial.set_axis_off()
    # if withFAC:
    axcbar = plt.subplot2grid((91, 3), (90, 1))
    axinfo = plt.subplot2grid((91, 3), (90, 0))
    axinfo.set_axis_off()

    tmpcpcpn = []
    tmpcpcps = []
    tmplocn, tmplocs = [],[]
    for i in range(3):
        for j in range(3):
            if i == j == 1:
                continue # skip
            _ca = ca[(i,j)]
            By, Bz = np.sin(_ca * d2r) * B, np.cos(_ca * d2r) * B
            if _ca in [0, -90, 180, 90]:
                axdial.text(np.sin(_ca * d2r), np.cos(_ca * d2r),
                            str(_ca) + r'$^\circ$',
                            # rotation = -_ca,
                            ha = 'center', va = 'center')
                axdial.plot([.75*np.sin(_ca * d2r), .85*np.sin(_ca * d2r)],
                            [.75*np.cos(_ca * d2r), .85*np.cos(_ca * d2r)],
                            color = 'lightgrey')

            if (i, j) == (1, 2):
                j = 1
            swipe_n.update_model(v,  By, Bz,  tilt, f107val)
            swipe_s.update_model(v, -By, Bz, -tilt, f107val)
            # swipe_n.update_model(-v, by, bz, dptilt, f107val)

            phin      = swipe_n.get_potential(mlat_n, mlt_n)
            phis      = swipe_s.get_potential(mlat_s, mlt_s)

            # add a constant to both potentials to make it easier to compare
            # half  = (phin.max() - phin.min()) / 2
            # phis = phis - phis.min() - half
            # phin = phin - phin.min() - half

            # reverse sign of phis to facilitate comparison
            # phis = -phis

            dPhiN = phin[OKindsn].max()-phin[OKindsn].min()
            dPhiS = phis[OKindss].max()-phis[OKindss].min()
            minn,maxn = np.argmin(phin[OKindsn]),np.argmax(phin[OKindsn])
            mins,maxs = np.argmin(phis[OKindss]),np.argmax(phis[OKindss])
    
            tmpcpcpn.append(dPhiN)
            tmpcpcps.append(dPhiS)
            tmplocn.append([minn,maxn])
            tmplocs.append([mins,maxs])

            axes[i][j].contourf(mlat_n, mlt_n, phin, levels = potlevels, cmap = plt.cm.bwr, extend = 'both')
            if not no_SH:
                axes[i][j].contour (mlat_s, mlt_s, phis, levels = potlevels, colors = 'k' , extend = 'both', linewidths = .4)

                showstring = f'$\Delta \Phi_N=$ {dPhiN:.0f}' + f' kV\n$\Delta \Phi_S=$ {dPhiS:.0f}' + f' kV\n$\Delta \Phi_N/\Delta \Phi_S=$ {dPhiN/dPhiS:.2f}'

            else:

                showstring = f'$\Delta \Phi_N=$ {dPhiN:.0f} kV'

            if doCalcBhattacharyya:
                # Separate into pos and neg current distributions
                psipos_n, psineg_n = np.copy(phin),-phin
                psipos_n[psipos_n < 0] = 0
                psineg_n[psineg_n < 0] = 0
            
                psipos_s, psineg_s = np.copy(phis),-phis
                psipos_s[psipos_s < 0] = 0
                psineg_s[psineg_s < 0] = 0
                
                phin_sum = np.sum(psipos_n*dS+psineg_n*dS)
                phis_sum = np.sum(psipos_s*dS+psineg_s*dS)
            
                # Normalize pos and neg current distributions
                psipos_n, psineg_n = sep_into_pos_and_neg_dist_and_norm(phin,dS)
                psipos_s, psineg_s = sep_into_pos_and_neg_dist_and_norm(phis,dS)
            
                # Calculate integrands for Bhattacharya coefficient
                facposintegrand = np.sqrt(psipos_n*psipos_s)
                facnegintegrand = np.sqrt(psineg_n*psineg_s)
                
                facBhatcoeff = np.sum(facposintegrand*dS+facnegintegrand*dS)

                showstring += "\n"+r"$BC = $"+f"{facBhatcoeff:.2f}"

            axes[i][j].write(55, 10.2, showstring,
                             ha = 'left', va = 'top', size = 9, multialignment='left')

    axdial.text(0, 0, u'IONOSPHERIC\nPOTENTIAL', ha = 'center', va = 'center', size = 14)

    plt.subplots_adjust(hspace = .01, wspace = .01, left = .01, right = .99, bottom = .05, top = .99)

    axdial.set_xlim(-1.2, 1.2)
    axdial.set_ylim(-1.2, 1.2)

    axcbar.contourf(potlevels, [0, 1], np.vstack((potlevels, potlevels)), levels = potlevels, cmap = plt.cm.bwr)
    axcbar.set_yticks([])
    axcbar.set_xlabel('kV')

    # Save cpcp info
    cpcpn.append(np.array(tmpcpcpn))
    cpcps.append(np.array(tmpcpcps))
    locns.append(np.array(tmplocn))
    locss.append(np.array(tmplocn))

    if np.isclose(tilt,0):
        tiltsign = ''
    elif tilt > 0:
        tiltsign = "±"
    else:
        tiltsign = "∓"
    axinfo.text(axinfo.get_xlim()[0], axinfo.get_ylim()[0], '$B_T$ = %s nT, $v$ = %s km/s,\n F$_{10.7}$ = %s, TILT $= %s%s^\circ$' % (B, v, f107val, tiltsign, np.abs(tilt)), ha = 'left', va = 'top', size = 12)

    if tilt >= 0:
        tiltstr = f"p{tilt}"
    else:
        tiltstr = f"n{np.abs(tilt)}"

    if no_SH:
        figname = plotdir+'swipe_potential_ns_comparison_' + tiltstr + '_deg_tilt_noSH.png'
    else:
        figname = plotdir+'swipe_potential_ns_comparison_' + tiltstr + '_deg_tilt.png'
    # if switchem:
    #     figname = figname.replace('_ns','_sn')
    #     if maskhemi == 0:
    #         figname = figname.replace("maskNH","maskSH")
    #     elif maskhemi == 1:
    #         figname = figname.replace("maskSH","maskNH")
    plt.savefig(figname, dpi = 300)

    # plt.savefig(plotdir+'swipe_potential_ns_comparison_' + str(tilt) + '_deg_tilt.png', dpi = 300)
    # plt.savefig(plotdir+'swipe_potential_ns_comparison_' + str(tilt) + '_deg_tilt.pdf')
    # plt.savefig(plotdir+'swipe_potential_ns_comparison_' + str(tilt) + '_deg_tilt.svg')

    # break

print("Save data ...")
plt.show()



import numpy as np
import pandas as pd
import os
from functools import reduce
from .sh_utils import get_A_matrix__Ephizero, get_A_matrix__potzero, SHkeys
from .constants import REFRE,d2r

basepath = os.path.dirname(__file__)
# basepath = '/SPENCEdata/Research/database/SHEIC/matrices/10k_points/'

default_coeff_fn = os.path.abspath(os.path.join(basepath,'coefficients','SW_OPER_MIO_SWI_2E_00000000T000000_99999999T999999_0101.txt'))

# read coefficient file and store in pandas DataFrame - with column names from last row of header:
colnames = ([x for x in open(default_coeff_fn).readlines() if x.startswith('#')][-1][1:]).strip().split(' ') 

get_coeffs = lambda fn: pd.read_table(fn, skipinitialspace = True, comment = '#', sep = ' ', names = colnames, index_col = [0, 1])


# organize the coefficients in arrays that are used to calculate magnetic field values
names = ['const', 'sinca', 'cosca',
         'epsilon', 'epsilon_sinca', 'epsilon_cosca', 
         'tilt', 'tilt_sinca', 'tilt_cosca',
         'tilt_epsilon', 'tilt_epsilon_sinca', 'tilt_epsilon_cosca',
         'tau', 'tau_sinca', 'tau_cosca',
         'tilt_tau', 'tilt_tau_sinca', 'tilt_tau_cosca',
         'f107']


def get_truncation_levels(coeff_fn = default_coeff_fn):
    """ read model truncation levels from coefficient file 
        returns NT, MT (spherical harmonic degree (N) and order (M)
        for toroidal (T) field(?))
    """

    # read relevant line and split in words:
    words = [l for l in open(coeff_fn).readlines() if 'Spherical harmonic degree' in l][0].split(' ')

    # remove commas from each word
    words = [w.replace(',', '') for w in words]

    # pick out the truncation levels and convert to ints
    NT, MT = [int(num) for num in words if num.isdigit()]

    return NT, MT


def get_m_matrix(coeff_fn = default_coeff_fn):
    """ make matrix of model coefficients - used in get_E for fast calculations
        of model field time series along trajectory with changing input
    """
    coeffs = get_coeffs(coeff_fn)

    # m_matrix = np.array([np.hstack((coeffs.loc[:, 'tor_c_' + ss].dropna().values,
    #                                 coeffs.loc[:, 'tor_s_' + ss].dropna().values,
    #                                 coeffs.loc[:, 'pol_c_' + ss].dropna().values,
    #                                 coeffs.loc[:, 'pol_s_' + ss].dropna().values)) for ss in names]).T
    m_matrix = np.array([np.hstack((coeffs.loc[:, 'tor_c_' + ss].dropna().values,
                                    coeffs.loc[:, 'tor_s_' + ss].dropna().values)) for ss in names]).T
    return m_matrix


def get_model_vectors(v, By, Bz, tilt, f107, epsilon_multiplier = 1., coeff_fn = default_coeff_fn):
    """ tor_c, tor_s = get_model_vectors(v, By, Bz, tilt, F107, epsilon_multiplier = 1., coeffs = coeffs)

        returns column vectors ((K,1)-shaped) corresponding to the spherical harmonic coefficients of the toroidal(?)
        part, with _c and _s denoting cos and sin terms, respectively.

        This function is used by sheic.SHEIC class
    """

    coeffs = get_coeffs(coeff_fn)
    coeffs = coeffs.fillna(0)

    ca = np.arctan2(By, Bz)
    epsilon = np.abs(v)**(4/3.) * np.sqrt(By**2 + Bz**2)**(2/3.) * (np.sin(ca/2)**(8))**(1/3.) / 1000 * epsilon_multiplier # Newell coupling           
    tau     = np.abs(v)**(4/3.) * np.sqrt(By**2 + Bz**2)**(2/3.) * (np.cos(ca/2)**(8))**(1/3.) / 1000 # Newell coupling - inverse 

    # make a dict of the 19 external parameters, where the keys are postfixes in the column names of coeffs:
    external_params = {'const'             : 1                          ,                            
                       'sinca'             : 1              * np.sin(ca),
                       'cosca'             : 1              * np.cos(ca),
                       'epsilon'           : epsilon                    ,
                       'epsilon_sinca'     : epsilon        * np.sin(ca),
                       'epsilon_cosca'     : epsilon        * np.cos(ca),
                       'tilt'              : tilt                       ,
                       'tilt_sinca'        : tilt           * np.sin(ca),
                       'tilt_cosca'        : tilt           * np.cos(ca),
                       'tilt_epsilon'      : tilt * epsilon             ,
                       'tilt_epsilon_sinca': tilt * epsilon * np.sin(ca),
                       'tilt_epsilon_cosca': tilt * epsilon * np.cos(ca),
                       'tau'               : tau                        ,
                       'tau_sinca'         : tau            * np.sin(ca),
                       'tau_cosca'         : tau            * np.cos(ca),
                       'tilt_tau'          : tilt * tau                 ,
                       'tilt_tau_sinca'    : tilt * tau     * np.sin(ca),
                       'tilt_tau_cosca'    : tilt * tau     * np.cos(ca),
                       'f107'              : f107                        }

    # The SH coefficients are the sums in the expansion in terms of external parameters, scaled by the ext. params.:
    tor_c = reduce(lambda x, y: x+y, [coeffs.loc[:, 'tor_c_' + param] * external_params[param] for param in external_params.keys()]).fillna(0)
    tor_s = reduce(lambda x, y: x+y, [coeffs.loc[:, 'tor_s_' + param] * external_params[param] for param in external_params.keys()]).fillna(0)
    # pol_c = reduce(lambda x, y: x+y, [coeffs.loc[:, 'pol_c_' + param] * external_params[param] for param in external_params.keys()]).dropna()
    # pol_s = reduce(lambda x, y: x+y, [coeffs.loc[:, 'pol_s_' + param] * external_params[param] for param in external_params.keys()]).fillna(0)
    # pol_s = pol_s.loc[pol_c.index] # equal number of sin and cos terms, but sin coeffs will be 0 where m = 0
    tor_s = tor_s.loc[tor_c.index] # 

    # ORIG
    # return tor_c.values[:, np.newaxis], tor_s.values[:, np.newaxis], tor_c.index.values

    # NEW
    tor_c, tor_s, tor_keys = tor_c.values[:, np.newaxis], tor_s.values[:, np.newaxis], tor_c.index.values

    keys_T = [c for c in tor_keys]
    m_T = np.array(keys_T).T[1][np.newaxis, :]
    n_T = np.array(keys_T).T[0][np.newaxis, :]

    N, M = np.max( np.hstack((np.array([c for c in tor_keys]).T, np.array([c for c in tor_keys]).T)), axis = 1)

    # nmax = np.max([key[0] for key in tor_keys])
    n_m0 = np.sum([key[1] == 0 for key in tor_keys])
    n_m1 = np.sum([key[1] == 1 for key in tor_keys])
    # n_m2 = np.sum([key[1] == 2 for key in tor_keys])
    # mmax = np.max([key[1] for key in tor_keys])

    # Do we need to apply A matrix to get remaining coefficients?
    apply_A = False
    if n_m1 < n_m0:
        apply_A = True

        A = get_A_matrix__Ephizero(N, M,
                                   zero_thetas = 90.-np.array([47.,-47.]),
                                   return_all = False)

        checkemout_coeffs = (n_T.ravel() < (N-1)) | (m_T.ravel() == 0)


    elif (n_m1 == n_m0) and (n_T.min() == 3):
        apply_A = True

        # warnings.warn("Zero thetas flipped!")
        # A = get_A_matrix__potzero(N, M,
        #                           zero_thetas = 90.-np.array([-47.,47.]),
        #                           return_all = False)

        A = get_A_matrix__potzero(N, M,
                                  zero_thetas = 90.-np.array([47.,-47.]),
                                  return_all = False)

        checkemout_coeffs = []
        for count,(n,m) in enumerate(zip(n_T.ravel(),m_T.ravel())):
            nprime = np.maximum(1,m)
            checkemout_coeffs.append( n >= (nprime + 2) )
        checkemout_coeffs = np.array(checkemout_coeffs)


    # Apply A, if necessary
    if apply_A:
        orig_tor_c = tor_c[checkemout_coeffs]
        orig_tor_s = tor_s[checkemout_coeffs]

        tmp_tor_c = orig_tor_c.copy()
        tmp_tor_s = orig_tor_s.copy()

        tor_c = A@tmp_tor_c
        tor_s = A@tmp_tor_s

        tor_keys = SHkeys(N, M).setNmin(1).MleN().Mge(0)

    return tor_c, tor_s, tor_keys


def get_apex_base_vectors(alat, height, h_R=110.):


    """ Calculate apex coordinate base vectors d_i and e_i (i = 1, 2, 3)

    The base vectors are defined in Richmond (1995). They can be calculated analytically
    for a dipole magnetic field and spherical Earth. 

    The output vectors will have shape (3, N) where N is the combined size of the input,
    after numpy broadcasting. The three rows correspond to east, north and radial components
    for a spherical Earth

    Note
    ----
    This function only calculates Modified Apex base vectors. QD base vectors f_i and g_i are 
    just eastward, northward, and radial unit vectors for a dipole field (and f_i = g_i)
        
    Parameters
    ----------
    height : array-like
        altitude of the points where the base vectors shall be calculated, in same unit as h_R
    alat : array-like
        Modified Apex(h_R) magnetic latitude, for reference height h_R  (degrees)
    h_R : float, optional
        Reference altitude used in modified apex coordinates (assuming Dipole magnetic field!). Default is 110 km

    Returns
    -------
    d1 : array-like
        modified apex base vector d1 for a dipole magnetic field, shape (3, N)
    d2 : array-like
        modified apex base vector d2 for a dipole magnetic field, shape (3, N)
    d3 : array-like
        modified apex base vector d3 for a dipole magnetic field, shape (3, N)
    e1 : array-like
        modified apex base vector e1 for a dipole magnetic field, shape (3, N)
    e2 : array-like
        modified apex base vector e2 for a dipole magnetic field, shape (3, N)
    e3 : array-like
        modified apex base vector e3 for a dipole magnetic field, shape (3, N)
    """
    try:
        h, ala = map(np.ravel, np.broadcast_arrays(height, alat))
    except:
        raise ValueError('get_apex_base_vectors: Input not broadcastable')

    R   = REFRE + h_R
    r   = REFRE + h

    ala = np.deg2rad(ala)

    N = r.size

    def alat2qdlat_dip(alatrad,r,R):
        sign = np.ones_like(alatrad)
        sign[alatrad >= 0] = 1.
        sign[alatrad <  0] = -1.

        qlatrad = np.arccos(np.sqrt((r/R)*np.cos(alatrad)**2))

        return sign * qlatrad

    la = alat2qdlat_dip(ala, r, R)

    if np.any(r / np.cos(la)**2 < R):
        raise ValueError('get_apex_base_vectors: Some points have apex height < R. Apex height is r/cos(lat)**2.')

    R2r = R / r 
    C   = np.sqrt(4 - 3 * R2r * np.cos(la)**2)
    
    d1 =  R2r ** (3./2) * np.vstack((np.ones(N), np.zeros(N), np.zeros(N)))
    d2 = -R2r ** (3./2) / C * np.vstack((np.zeros(N), 2 * np.sin(la), np.cos(la)))
    d3 =  R2r ** (-3) * C / (4 - 3 * np.cos(la)**2) * np.vstack((np.zeros(N), np.cos(la), -2*np.sin(la)))

    e1 = np.cross(d2.T, d3.T).T
    e2 = np.cross(d3.T, d1.T).T
    e3 = np.cross(d1.T, d2.T).T

    return (d1, d2, d3, e1, e2, e3)




# IPython log file

# import swipe
import numpy as np
from pyswipe.swipe import SWIPE

plotdir = '/plotdir/'

import matplotlib as mpl
mplBkgrnd = 'QtAgg'
mpl.use(mplBkgrnd)

# MPL opts

mpl.rcParams.update({'text.color': 'k'})
mpl.rcParams.update({'axes.labelcolor': 'k'})
mpl.rcParams.update({'xtick.color': 'k'})
mpl.rcParams.update({'ytick.color': 'k'})
mpl.rcParams.update({'font.size': 15})
# mpl.rcParams.update({'font.family': 'sans-serif'})
# mpl.rcParams.update({'font.sans-serif': 'Arial'})
mpl.rcParams.update({'text.usetex': False})

mpl.rcParams.update({'figure.figsize': [10.0, 8.0]})
mpl.rcParams.update({'savefig.directory': plotdir})

import matplotlib.pyplot as plt
plt.ion()

##############################
get_var_str = 'convection'
# get_var_str = 'efield'

bys = np.arange(-5,5.1,5.)
bz = -5

swvel = 360
f107val = 120

savefig = False

##############################

thiss = []
potss = []
zero_latNU = None

minlat = 45
minlat_for_cpcp_calc = 70

# bzs = [-3.,0.,3.]
# bzs = [-3., 0., 3.]
dptilts = np.arange(-30,30.1,15.)
dptilts = np.arange(-20,20.1,20.)

Edn = dict()                    # E-field dictionary, north
Eds = dict()
# for i in range(len(coefffiles)):
i = 0

# lists for holding E-field for each bz value
ENss, ESss = [], []
for k,dptilt in enumerate(dptilts):

    # lists for holding CPCP in each hemisphere and the locations of extrema
    ENs, ESs = [], []
    for j,by in enumerate(bys):

        print(i,k,j,by,bz,dptilt)

        if j == 0:
            this = SWIPE(-swvel,by,bz,dptilt,f107val,
                         minlat=minlat,
                         zero_lats=zero_latNU)

            mlats, mlts = this.scalargrid
            mlatv, mltv = this.vectorgrid
            mlatvn, mlatvs = np.split(mlatv, 2)
            mltvn, mltvs = np.split(mltv, 2)
            mlatsn, mlatss = np.split(mlats, 2)
            mltsn, mltss = np.split(mlts, 2)

            sizes = int(np.sqrt(np.split(mlats,2)[0].size))  # size of scalar grid

            # reshaper for scalar grid
            def reshapes(q):
                qn, qs = np.split(q,2)
                qn, qs = qn.reshape(sizes,sizes), qs.reshape(sizes,sizes)
                return qn, qs

            # lat-averaged stats 
            def latstats(qh):
                """'qh' means 'quantity in a particular hemisphere', 
                so q after running it through 'reshapes'
                """
                return np.mean(qh,axis=0), np.std(qh,axis=0)

            mlatsnu, mlatssu = mlatsn.reshape(100,100)[0,:], mlatss.reshape(100,100)[0,:]

        else:
            this.update_model(-swvel, by, bz, dptilt, f107val)

        # phin, phis = np.split(this.get_potential(), 2)
        # E1, E2: E_d1 and E_d2 field-perp components in Modified Apex coordinates
        if get_var_str == 'efield':
            E1, E2 = this.get_efield_MA(mlat=mlats, mlt=mlts,return_magnitude=False)
        elif get_var_str == 'convection':
            E1, E2, Emag = this.get_convection_vel_MA(mlat=mlats, mlt=mlts,return_magnitude=True)

        E1n, E1s = reshapes(E1)
        E2n, E2s = reshapes(E2)

        E1nmean, E1nstd = latstats(E1n)
        E2nmean, E2nstd = latstats(E2n)

        E1smean, E1sstd = latstats(E1s)
        E2smean, E2sstd = latstats(E2s)

        if get_var_str == 'convection':
            Emagn, Emags = reshapes(Emag)
            Emagnmean, Emagnstd = latstats(Emagn)
            Emagsmean, Emagsstd = latstats(Emags)

            ENs.append({1:[E1nmean,E1nstd],2:[E2nmean,E2nstd],'mag':[Emagnmean,Emagnstd]})
            ESs.append({1:[E1smean,E1sstd],2:[E2smean,E2sstd],'mag':[Emagsmean,Emagsstd]})
        else:
            ENs.append({1:[E1nmean,E1nstd],2:[E2nmean,E2nstd]})
            ESs.append({1:[E1smean,E1sstd],2:[E2smean,E2sstd]})

    ENss.append(ENs)
    ESss.append(ESs)

Edn[i] = ENss
Eds[i] = ESss

########################################

cmap = 'magma'
cmap2 = 'bwr'
labs = ['NH','SH',r"NH$-$SH"]

if get_var_str == 'efield':
    unitstr = 'mV/m'
    l1 = r"$E_{d1}$ (E-W)"
    l2 = r"$E_{d2}$ (N-S)"
elif get_var_str == 'convection':
    unitstr = 'm/s'
    l1 = r"$v_{e1}$ (E-W)"
    l2 = r"$v_{e2}$ (N-S)"
    l3 = r"$v_{\mathrm{mag}}$"

tilti = 0
byi = 2

tmptilt, tmpby = dptilts[tilti], bys[byi]

# E1nmean, E1nstd = Edn[0][tilti][byi][1]
# E2nmean, E2nstd = Edn[0][tilti][byi][2]

# E1smean, E1sstd = Eds[0][tilti][byi][1]
# E2smean, E2sstd = Eds[0][tilti][byi][2]

fig = plt.figure(9,figsize = (10, 10) if get_var_str == 'efield' else (15,10))
fig.clf()
tottitl = f"$B_y$={tmpby:.0f}, $B_z$={bz:.0f}"
fig.suptitle(tottitl)
if get_var_str == 'convection':
    ax1 = plt.subplot(331)
    ax2 = plt.subplot(332)
    ax3 = plt.subplot(333)
    ax4 = plt.subplot(334)
    ax5 = plt.subplot(335)
    ax6 = plt.subplot(336)
    ax7 = plt.subplot(337)
    ax8 = plt.subplot(338)
    ax9 = plt.subplot(339)

    axrows = [[ax1,ax2,ax3],[ax4,ax5,ax6],[ax7,ax8,ax9]]

elif get_var_str == 'efield':
    ax1 = plt.subplot(321)
    ax2 = plt.subplot(322)
    ax3 = plt.subplot(323)
    ax4 = plt.subplot(324)
    ax5 = plt.subplot(325)
    ax6 = plt.subplot(326)
    
    axrows = [[ax1,ax2],[ax3,ax4],[ax5,ax6]]

stri1 = slice(None,None,4)
stri2 = slice(2,None,4)

tilt_i = [(2,0),(1,1),(0,2)]    # Assumes dptilts = [-15,0,15]. Local summer first, then equinox, then local winter

for i in range(len(dptilts)):

    ni,si = tilt_i[i]
    E1nmean, E1nstd = Edn[0][ni][byi][1]
    E2nmean, E2nstd = Edn[0][si][byi][2]

    E1smean, E1sstd = Eds[0][ni][byi][1]
    E2smean, E2sstd = Eds[0][si][byi][2]

    axrow = axrows[i]
    
    tmptilt = dptilts[i]

    plt.sca(axrow[0])

    plt.plot(mlatsnu        ,E1nmean,color='blue',label='NH')
    plt.plot(np.abs(mlatssu),E1smean,color='red' ,label='SH')
    plt.axvline(47,color='gray',linestyle=':')
    plt.fill_between(mlatsnu        ,E1nmean-E1nstd,E1nmean+E1nstd,color='blue',alpha=0.2,hatch='\\')
    plt.fill_between(np.abs(mlatssu),E1smean-E1sstd,E1smean+E1sstd,color='red' ,alpha=0.2,hatch='/')
    if i == 0:
        plt.title(l1)
        plt.legend(loc='upper left')
    
    plt.ylabel("["+unitstr+"]")
    if i == 2:
        plt.xlabel("MA Latitude [deg]")

    plt.sca(axrow[1])
    plt.plot(mlatsnu        ,E2nmean,color='blue',label='NH')
    plt.plot(np.abs(mlatssu),E2smean,color='red' ,label='SH')
    plt.axvline(47,color='gray',linestyle=':')
    plt.fill_between(mlatsnu        ,E2nmean-E2nstd,E2nmean+E2nstd,color='blue',alpha=0.2,hatch='\\')
    plt.fill_between(np.abs(mlatssu),E2smean-E2sstd,E2smean+E2sstd,color='red' ,alpha=0.2,hatch='/')
    if i == 0:
        plt.title(l2)

    if i == 2:
        plt.xlabel("MA Latitude [deg]")

    if get_var_str == 'convection':
        
        Emagnmean, Emagnstd = Edn[0][ni][byi]['mag']
        Emagsmean, Emagsstd = Eds[0][ni][byi]['mag']

        plt.sca(axrow[2])
        plt.plot(mlatsnu        ,Emagnmean,color='blue',label='NH')
        plt.plot(np.abs(mlatssu),Emagsmean,color='red' ,label='SH')
        plt.axvline(47,color='gray',linestyle=':')
        plt.fill_between(mlatsnu        ,Emagnmean-Emagnstd,Emagnmean+Emagnstd,color='blue',alpha=0.2,hatch='\\')
        plt.fill_between(np.abs(mlatssu),Emagsmean-Emagsstd,Emagsmean+Emagsstd,color='red' ,alpha=0.2,hatch='/')
        if i == 0:
            plt.title(l3)

tmp3 = fig.text(0.03,0.765,"Summer", size = 18, rotation='vertical', va='center')
tmp2 = fig.text(0.03,0.49,"Equinox", size = 18, rotation='vertical', va='center')
tmp1 = fig.text(0.03,0.22,"Winter" , size = 18, rotation='vertical', va='center')
plt.subplots_adjust(left=0.15, bottom=0.1, right=0.95, top=0.9, wspace=0.2, hspace=None)

if savefig:
    fname = "16__lat_averages/"+get_var_str+'_by_season_'+tottitl.replace('$','').replace('_','').replace(' ','').replace('$^\circ$','deg')+'.png'
    fname = fname.replace("\\Psi","ψ").replace("^\\circ","deg").replace("\n","_").replace("{SW}","sw").replace("[","_").replace("]","").replace('CPCP_','')
    print("Saving fig to "+fname)
    plt.savefig(plotdir+fname,dpi=300)
        
else:
    print("NOT saving plot")



from __future__ import absolute_import
from .swipe import SWIPE
from .swipe import get_v, get_E, get_pflux
from .mlt_utils import mlon_to_mlt
import pyswipe.plot_utils
import pyswipe.sh_utils
import pyswipe.model_utils


__all__ = ["SWIPE","get_B_ground","get_B_space","mlon_to_mlt"]

__version__ = "0.1"



""" 
This module contains tools for plotting certain quantities in polar coordinates
(mlt/mlat grid).

Example:
--------
import matplotlib.pyplot as plt
import numpy as np
fig = plt.figure()
ax = fig.add_subplot(111)   
pax = Polarsubplot(ax)
pax.plot(np.ones(20) * 60, np.linspace(18, 24+6, 20), color = 'red', linewidth = 5)
pax.scatter([70, 70], [15, 9], s = 200, c = 'red')
plt.show()

In addition to `scatter` and `plot`, it is also possible to use matplotlib's 
`contour` and `contourf` functions with mlt/mlat coordinates. A function 
`featherplot` is also added for plotting vectors. A function `write` is also 
provided which replicates the beahaviour of matplotlib's `text`.



MIT License

Copyright (c) 2017 Karl M. Laundal

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


from __future__ import absolute_import, division
import numpy as np
from scipy.interpolate import griddata
from builtins import range
import matplotlib.pyplot as plt
import warnings
from matplotlib import rc
from matplotlib.patches import Polygon, Ellipse
from matplotlib.collections import PolyCollection, LineCollection


d2r = np.pi / 180

rc('font', **{'family': 'sans-serif', 'sans-serif': ['Verdana']})
rc('text', usetex=True)

class Polarplot(object):
    def __init__(self, ax, minlat = 50, plotgrid = True, sector = 'all', lt_label='lt', lat_label='lat', **kwargs):
        """ pax = Polarsubplot(axis, minlat = 50, plotgrid = True, **kwargs)

            **kwargs are the plot parameters for the grid

            this is a class which handles plotting in polar coordinates, specifically a latitude-local time grid

            Example:
            --------
            import matplotlib.pyplot as plt
            fig = plt.figure()
            ax = fig.add_subplot(111)
            pax = Polarsubplot(ax)
            pax.MEMBERFUNCTION()
            plt.show()


            where memberfunctions include:
            plotgrid()                           - called by __init__
            plot(lat, lt, **kwargs)            - works like plt.plot
            text(lat, lt, text, **kwargs)      - works like plt.text
            scatter(lat, lt, **kwargs)         - works like plt.scatter
            contour(lat, lt, f)                - works like plt.contour
            contourf(lat, lt, f)               - works like plt.contourf


        Parameters
        ----------
        ax : matplotlib AxesSubplot object
        minlat : scalar, optional
            low latitude boundary of the plot. The default is 50.
        plotgrid : bool, optional
            Set True to plot a grid. The default is True.
        sector : string, optional
            Used to generate portions of polar plot.
            Can either use one of the following strings: 'all', 'dusk', 'dawn', 'night', 'day'. Or can be defined using numbers for example '18-6' will produce the same as night while '6-18' will produce the same as day. The default is 'all'.
        lt_label : string, optional
            Name of the local time co-ordinate used for displaying the hover co-ordinates. The default is 'lt'.
        lat_label : string, optional
            Name of the latitude co-ordinate used for displaying the hover co-ordinates. The defualt is 'lat'.
        **kwargs : dict
            Keywords passed to the plot function to control grid lines.

        """
        self.minlat = minlat # the lower latitude boundary of the plot
        self.ax = ax
        self.ax.axis('equal')
        self.sector = sector

        if 'linewidth' not in kwargs.keys():
            kwargs['linewidth'] = .5

        if 'color' not in kwargs.keys():
            kwargs['color'] = 'lightgrey'

        if 'linestyle' not in kwargs.keys():
            kwargs['linestyle'] = '--'

        # define the lt limits
        if sector.lower() == 'all':
            self.ltlims = 'lt >= 0'
        elif sector.lower() == 'dusk':
            self.ltlims = '(lt >= 12) | (lt == 0)'
        elif sector.lower() == 'dawn':
            self.ltlims = '(lt <= 12) | (lt == 24)'
        elif sector.lower() == 'night':
            self.ltlims = '(lt >= 18) | (lt <= 6)'
        elif sector.lower() == 'day':
            self.ltlims = '(lt >= 6) & (lt <=18)'
        else:
            sector = [float(s) for s in sector.split('-')]
            if sector[0]> sector[-1]:
                self.ltlims= f'(lt>={sector[0]})|(lt<={sector[-1]})'
            else:
                self.ltlims= f'(lt>={sector[0]})&(lt<={sector[-1]})'

        self.ax.set_axis_off()

        self.ax.format_coord= self.make_format(lt_label, lat_label)
        if plotgrid:
            self.plotgrid(**kwargs)

        # set suitable plot limits by drawing a circle at minlat:
        x, y = self._latlt2xy(np.full(100, self.minlat), np.linspace(0, 24, 100))
        x, y = np.hstack((x, 0)), np.hstack((y, 0)) # add origin
        self.ax.set_xlim(np.nanmin(x) - 0.1, np.nanmax(x) + 0.1)
        self.ax.set_ylim(np.nanmin(y) - 0.1, np.nanmax(y) + 0.1)


    def plot(self, lat, lt, **kwargs):
        """
        Wrapper for matplotlib's plot function. Accepts lat, lt instead of x and y

        keywords passed to this function is passed on to matplotlib's plot
        """

        x, y = self._latlt2xy(lat, lt)
        return self.ax.plot(x, y, **kwargs)


    def text(self, lat, lt, text, ignore_plot_limits=False, **kwargs):
        """
        Wrapper for matplotlib's text function. Accepts lat, lt instead of x and y

        keywords passed to this function is passed on to matplotlib's text
        """

        x, y = self._latlt2xy(lat, lt, ignore_plot_limits=ignore_plot_limits)

        if np.isfinite(x + y):
            return self.ax.text(x, y, text, **kwargs)
        else:
            print('text outside plot limit - set "ignore_plot_limits = True" to override')


    def write(self, lat, lt, text, ignore_plot_limits=False, **kwargs):
        """ Alias for text, a wapper for matplotlib's text function. Accepts lat, lt instead of x and y

        keywords passed to this function is passed on to matplotlib's text
        """

        return self.text(lat, lt, text, ignore_plot_limits, **kwargs)


    def scatter(self, lat, lt, **kwargs):
        """
        Wrapper for matplotlib's scatter function. Accepts lat, lt instead of x and y

        keywords passed to this function is passed on to matplotlib's scatter
        """

        x, y = self._latlt2xy(lat, lt)
        c = self.ax.scatter(x, y, **kwargs)
        return c


    def plotgrid(self, labels=False, **kwargs):
        """ plot lt, lat-grid on self.ax

        parameters
        ----------
        labels: bool
            set to True to include lat/lt labels
        **kwarsgs: dictionary
            passed to matplotlib's plot function (for linestyle etc.)

        """
        returns= []
        for lt in [0, 6, 12, 18]:
            returns.append(self.plot([self.minlat, 90], [lt, lt], **kwargs))

        lts = np.linspace(0, 24, 100)
        for lat in np.r_[90: self.minlat -1e-12 :-10]:
            returns.append(self.plot(np.full(100, lat), lts, **kwargs))

        # add LAT and LT labels to axis
        if labels:
            returns.append(self.writeLATlabels())
            returns.append(self.writeLTlabels())
        return tuple(returns)

    def writeLTlabels(self, lat = None, degrees = False, **kwargs):
        """ write local time labels at given latitude (default minlat - 2)
            if degrees is true, the longitude will be written instead of hour (with 0 at midnight)
        """
        if lat is None:
            lat = self.minlat - 2
        labels=[]
        if degrees:
            if self.sector in ['all', 'night', 'dawn', 'dusk']:
                labels.append(self.write(lat, 0,    '0$^\circ$', verticalalignment = 'top'   , horizontalalignment = 'center', ignore_plot_limits=True, **kwargs))
            if self.sector in ['all', 'night', 'dawn', 'day']:
                labels.append(self.write(lat, 6,   '90$^\circ$', verticalalignment = 'center', horizontalalignment = 'left'  , ignore_plot_limits=True, **kwargs))
            if self.sector in ['all', 'dusk', 'dawn', 'day']:
                labels.append(self.write(lat, 12, '180$^\circ$', verticalalignment = 'bottom', horizontalalignment = 'center', ignore_plot_limits=True, **kwargs))
            if self.sector in ['all', 'night', 'dusk', 'day']:
                labels.append(self.write(lat, 18, '-90$^\circ$', verticalalignment = 'center', horizontalalignment = 'right' , ignore_plot_limits=True, **kwargs))
        else:
            lt=np.array([0, 24])
            if any(eval(self.ltlims)):
                labels.append(self.write(lat, 0, '00', verticalalignment = 'top'    , horizontalalignment = 'center', ignore_plot_limits=True, **kwargs))
            lt=6
            if eval(self.ltlims):
                labels.append(self.write(lat, 6, '06', verticalalignment = 'center' , horizontalalignment = 'left'  , ignore_plot_limits=True, **kwargs))
            lt=12
            if eval(self.ltlims):
                labels.append(self.write(lat, 12, '12', verticalalignment = 'bottom', horizontalalignment = 'center', ignore_plot_limits=True, **kwargs))
            lt=18
            if eval(self.ltlims):
                labels.append(self.write(lat, 18, '18', verticalalignment = 'center', horizontalalignment = 'right' , ignore_plot_limits=True, **kwargs))

            return labels


    # added by AØH 20/06/2022 to plot latitude labels
    def writeLATlabels(self, lt = None, **kwargs):
        """ write latitude labels """
        if lt == None:
            lt = 3
        if kwargs is not None:
            latkwargs = {'rotation':45, 'color':'lightgrey', 'backgroundcolor':'white', 'zorder':2, 'alpha':1.}
            latkwargs.update(kwargs)
        labels = []
        for lat in np.r_[self.minlat:81:10]:
            labels.append(self.write(lat, lt, str(lat)+'$^{\circ}$', ignore_plot_limits = False, **latkwargs))

        return labels


    def plotpins(self, lats, lts, north, east, rotation = 0, SCALE = None, size = 10, unit = '', colors = 'black', markercolor = 'black', marker = 'o', markersize = 20, **kwargs):
        """ plot vector field. Each vector is a dot with a line pointing in the vector direction

            kwargs go to ax.plot

            the markers at each pin can be modified by the following keywords, that go to ax.scatter:
            marker (default 'o')
            markersize (defult 20 - size in points^2)
            markercolor (default black)

        """
        returns= []
        lts = lts.flatten()
        lats = lats.flatten()
        north = north.flatten()
        east = east.flatten()
        R = np.array(([[np.cos(rotation), -np.sin(rotation)], [np.sin(rotation), np.cos(rotation)]]))
        
        if SCALE is None:
            scale = 1.
        else:

            if unit is not None:
                returns.append(self.ax.plot([0.9, 1], [0.95, 0.95], color = colors, linestyle = '-', linewidth = 2))
                returns.append(self.ax.text(0.9, 0.95, ('%.1f ' + unit) % SCALE, horizontalalignment = 'right', verticalalignment = 'center', size = size))

            scale = 0.1/SCALE


        x, y = self._latlt2xy(lats, lts)
        dx, dy = R.dot(self._northEastToCartesian(north, east, lts))
        segments = np.dstack((np.vstack((x, x + dx * scale)).T, np.vstack((y, y + dy * scale)).T))


        returns.append(self.ax.add_collection(LineCollection(segments, colors = colors, **kwargs)))

        if markersize != 0:
            returns.append(self.scatter(lats, lts, marker = marker, c = markercolor, s = markersize, edgecolors = markercolor))
        return tuple(returns)

    def quiver(self, lat, lt, north, east, rotation = 0, qkeyproperties = None, **kwargs):
        """ wrapper for matplotlib's quiver function, just for lat/lt coordinates and
            east/north components. Rotation specifies a rotation angle for the vectors,
            which could be convenient if you want to rotate magnetic field measurements
            to an equivalent current direction

            rotation in radians

            qkeyproperties: set to dict to pass to matplotlib quiverkey. The dict must contain
            'label': the label written next to arrow
            'U': length of the arrow key in same unit as north/east components

        """

        lt = lt.flatten()
        lat = lat.flatten()
        north = north.flatten()
        east = east.flatten()
        R = np.array(([[np.cos(rotation), -np.sin(rotation)], [np.sin(rotation), np.cos(rotation)]]))

        x, y = self._latlt2xy(lat, lt)
        dx, dy = R.dot(self._northEastToCartesian(north, east, lt))

        q = self.ax.quiver(x, y, dx, dy, **kwargs)

        if qkeyproperties is not None:
            qkeykwargs = {'X':.8, 'Y':.9, 'labelpos':'E'}
            qkeykwargs.update(qkeyproperties)

            self.ax.quiverkey(q, **qkeykwargs)

        return q



    def plottrack(self, lats, lts, north, east, rotation = 0, SCALE = None, size = 10, unit = '', color = 'black', markercolor = 'black', marker = 'o', markersize = 20, **kwargs):
        """ like plotpins, only it's not a line pointing in the arrow direction but a dot at the tip of the arrow

            kwargs go to ax.plot

            the markers at each pin can be modified by the following keywords, that go to ax.scatter:
            marker (default 'o')
            markersize (defult 20 - size in points^2)
            markercolor (default black)

        """

        lts = lts.flatten()
        lats = lats.flatten()
        north = north.flatten()
        east = east.flatten()
        R = np.array(([[np.cos(rotation), -np.sin(rotation)], [np.sin(rotation), np.cos(rotation)]]))
        if 'key_coords' in kwargs:
            key_x, key_y= kwargs['key_coords']
            kwargs.pop('key_coords', None)
        else:
            key_x, key_y= 0.9, 0.95
        key=[]
        if SCALE is None:
            scale = 1.
        else:

            if unit is not None:
                key.append(self.ax.plot([key_x, key_x+.1], [key_y]*2, color = color, linestyle = '-', linewidth = 2))
                key.append(self.ax.text(key_x, key_y, ('%.1f ' + unit) % SCALE, horizontalalignment = 'right', verticalalignment = 'center', size = size))

            scale = 0.1/SCALE

        xs, ys, dxs, dys = [],[],[],[]
        for i in range(len(lats)):

            lt = lts[i]
            lat = lats[i]

            x, y = self._latlt2xy(lat, lt)
            dx, dy = R.dot(self._northEastToCartesian(north[i], east[i], lt).reshape((2, 1))).flatten()
            xs.append(x)
            ys.append(y)
            dxs.append(dx)
            dys.append(dy)

        xs, ys, dxs, dys = np.array(xs),np.array(ys),np.array(dxs)*scale,np.array(dys)*scale
        if len(key)!=0:
            return self.ax.scatter(xs+dxs,ys+dys,marker = marker, c = markercolor, s = markersize, edgecolors = markercolor), key
        else:
            return self.ax.scatter(xs+dxs,ys+dys,marker = marker, c = markercolor, s = markersize, edgecolors = markercolor)


    def contour(self, lat, lt, f, **kwargs):
        """
        Wrapper for matplotlib's contour function. Accepts lat, lt instead of x and y

        keywords passed to this function is passed on to matplotlib's contour
        """
        xea, yea = self._latlt2xy(lat.flatten(), lt.flatten(), ignore_plot_limits = True)
        mask, _  = self._latlt2xy(lat.flatten(), lt.flatten(), ignore_plot_limits = False)
        f = f.flatten()
        f[~np.isfinite(mask)] = np.nan

        # convert to cartesian uniform grid
        xx, yy = np.meshgrid(np.linspace(-1, 1, 150), np.linspace(-1, 1, 150))
        points = np.vstack( tuple((xea, yea)) ).T
        gridf = griddata(points, f, (xx, yy))

        # ... and plot
        return self.ax.contour(xx, yy, gridf, **kwargs)


    def contourf(self, lat, lt, f, **kwargs):
        """
        Wrapper for matplotlib's contourf function. Accepts lat, lt instead of x and y

        keywords passed to this function is passed on to matplotlib's contourf
        """

        xea, yea = self._latlt2xy(lat.flatten(), lt.flatten(), ignore_plot_limits = True)
        mask, _  = self._latlt2xy(lat.flatten(), lt.flatten(), ignore_plot_limits = False)
        f = f.flatten()
        f[~np.isfinite(mask)] = np.nan

        # convert to cartesian uniform grid
        xx, yy = np.meshgrid(np.linspace(-1, 1, 150), np.linspace(-1, 1, 150))
        points = np.vstack( tuple((xea, yea)) ).T
        gridf = griddata(points, f, (xx, yy))

        # ... and plot
        return self.ax.contourf(xx, yy, gridf, **kwargs)


    def fill(self, lat, lt, **kwargs):
        """ Fill polygon defined in lat/lt, **kwargs are given to self.ax.contour. LT in hours - no rotation
        """

        xx, yy = self._latlt2xy(lat.flatten(), lt.flatten())

        # plot
        return self.ax.fill(xx, yy, **kwargs)


    def plot_terminator(self, time, sza = 90, north = True, apex = None, shadecolor = None, shade_kwargs={}, **kwargs):
        """ shade the area antisunward of the terminator

            Parameters
            ----------
            time : float or datetime
                if datetime, time describes the time for which to calculate the
                subsolar point and associated sunlight terminator. If float, it
                desribes where to draw a horizontal terminator line - in
                degrees from the pole towards the night side (think of it
                approximately as dipole tilt angle)
            sza : float, optional
                the terminator is defined as contours of this solar zenith angle
                default 90 degrees
            north : bool, optional
                set to True for terminator in the north (default), False for south
            apex : apexpy.Apex object, optional
                give an apexpy.Apex object to convert the terminator to magnetic
                apex coordinates. By default it is plotted in geographic
            shadecolor : string, optional
                color of a shade to be drawn on the night side of the
                terminator. Default is None, which means that no shade will be
                drawn
            shade_kwargs : dict
                Used if shadecolor is not None. keywords passed to matplotlib.patches'
                Pylogon function
            kwargs : dict
                keywords passed to matplotlib's plot function for the terminator line
        """
        returns= []
        hemisphere = 1 if north else -1

        if np.isscalar(time): # time is interpreted as dipole tilt angle
            y0 = -(hemisphere * time + sza - 90) / (90. - self.minlat) # noon-midnight coordinate
            xr = np.sqrt(1 - y0**2)

            x = np.array([-xr, xr]).flatten()
            y = np.array([y0, y0]).flatten()

        else: # time should be a datetime object

            sslat, sslon = subsol(time)

            # make cartesian vector
            x = np.cos(sslat * d2r) * np.cos(sslon * d2r)
            y = np.cos(sslat * d2r) * np.sin(sslon * d2r)
            z = np.sin(sslat * d2r)
            ss = np.array([x, y, z]).flatten()

            # construct a vector pointing roughly towards dawn, and normalize
            t0 = np.cross(ss, np.array([0, 0, 1]))
            t0 = t0/np.linalg.norm(t0)

            # make a new vector pointing northward at the 90 degree SZA contour:
            sza90 = np.cross(t0, ss)

            # rotate this about the dawnward vector to get specified SZA contour
            rotation_angle = (sza - 90) * d2r

            sza_vector = sza90 * np.cos(rotation_angle) + np.cross(t0, sza90) * np.sin(rotation_angle) + t0 * (np.sum(t0*sza90)) * (1 - np.cos(rotation_angle)) # (rodrigues formula)
            sza_vector = sza_vector.flatten()

            # rotate this about the sun-Earth line to trace out the trajectory:
            angles = np.r_[-np.pi/2 : 3*np.pi/2: 2*np.pi / 360][np.newaxis, :]
            r = sza_vector[:, np.newaxis] * np.cos(angles) + np.cross(ss, sza_vector)[:, np.newaxis] * np.sin(angles) + ss[:, np.newaxis] * (np.sum(t0*sza90)) * (1 - np.cos(rotation_angle))

            # convert to spherical
            t_lat = 90 - np.arccos(r[2]) / d2r
            t_lon = (np.arctan2(r[1], r[0])*180/np.pi) % 360

            londiff = (t_lon - sslon + 180) % 360 - 180 # signed difference in longitude
            t_lt = (180. + londiff)/15. # convert to lt with ssqlon at noon

            if apex is not None:
                mlat, mlon = apex.geo2apex(t_lat, t_lon, apex.refh)
                mlt = apex.mlon2mlt(mlon, time)
                t_lat, t_lt = mlat, mlt

            # limit contour to correct hemisphere:
            iii = (t_lat >= self.minlat) if north else (t_lat <= -self.minlat)
            if len(iii) == 0:
                return 0 # terminator is outside plot
            t_lat = t_lat[iii] #* hemisphere
            t_lt = t_lt[iii]

            x, y = self._latlt2xy(t_lat, t_lt)


        returns.append(self.ax.plot(x, y, **kwargs))

        x0, x1 = np.min([x[0], x[-1]]), np.max([x[0], x[-1]])
        y0, y1 = np.min([y[0], y[-1]]), np.max([y[0], y[-1]])

        a0 = np.arctan2(y0, x0)
        a1 = np.arctan2(y1, x1)
        if a1 < a0: a1 += 2*np.pi

        a = np.linspace(a0, a1, 100)[::-1]

        xx, yy = np.cos(a), np.sin(a)

        if shadecolor is not None:
             shade = Polygon(np.vstack((np.hstack((x[::hemisphere], xx)), np.hstack((y[::hemisphere], yy)))).T, closed = True, color = shadecolor, linewidth = 0, **shade_kwargs)
             returns.append(self.ax.add_patch(shade))
        return tuple(returns)



    def filled_cells(self, lat, lt, latres, ltres, data, resolution = 10, crange = None, levels = None, bgcolor = None, verbose = False, **kwargs):
        """ specify a set of cells in lat and lt, along with a data array,
            and make a color plot of the cells

        """

        if not all([len(lat) == len(q) for q in [lt,ltres,data]]):
            print("WARNING: Input arrays to filled_cells of unequal length! Your plot is probably incorrect")

        lat, lt, latres, ltres, data = map(np.ravel, [lat, lt, latres, ltres, data])
        mask = np.ones_like(data)
        mask[(~eval(self.ltlims)) | (lat < self.minlat)] = np.nan
        data*=mask
        if verbose:
            print(lt.shape, lat.shape, latres.shape, ltres.shape)

        la = np.vstack([(lt - 6) / 12. * np.pi + i * ltres / (resolution - 1.) / 12. * np.pi for i in range(resolution)]).T
        if verbose:
            print (la.shape)
        ua = la[:, ::-1]

        vertslo = np.dstack(((90 - lat          )[:, np.newaxis] / (90. - self.minlat) * np.cos(la),
                             (90 - lat          )[:, np.newaxis] / (90. - self.minlat) * np.sin(la)))
        vertshi = np.dstack(((90 - lat - latres)[:, np.newaxis] / (90. - self.minlat) * np.cos(ua),
                             (90 - lat - latres)[:, np.newaxis] / (90. - self.minlat) * np.sin(ua)))
        verts = np.concatenate((vertslo, vertshi), axis = 1)

        if verbose:
            print( verts.shape, vertslo.shape, vertshi.shape)


        if 'cmap' in kwargs.keys():
            cmap = kwargs['cmap']
            kwargs.pop('cmap')
        else:
            cmap = plt.cm.viridis

        if levels is not None:
            # set up a function that maps data values to color levels:
            nlevels = len(levels)
            lmin, lmax = levels.min(), levels.max()
            self.colornorm = lambda x: plt.Normalize(lmin, lmax)(np.floor((x - lmin) / (lmax - lmin) * (nlevels - 1)) / (nlevels - 1) * (lmax - lmin) + lmin)
            coll = PolyCollection(verts, facecolors = cmap(self.colornorm(data.flatten())), **kwargs)

        else:
            coll = PolyCollection(verts, array = data, cmap = cmap, **kwargs)
            if crange is not None:
                coll.set_clim(crange[0], crange[1])



        if bgcolor != None:
            radius = 2*((90 - self.minlat)/ (90 - self.minlat))
            bg = Ellipse([0, 0], radius, radius, zorder = 0, facecolor = bgcolor)
            self.ax.add_artist(bg)


        return self.ax.add_collection(coll)


    def plotimg(self,lat,lt,image,corr=True , crange = None, bgcolor = None, **kwargs):
        """
        Displays an image in polar coordinates.
        lat,lt,image are m x n arrays.
        If corr == True, lat and lt are corrected to edge coordinates. The plotted image is (m-2) x (n-2)
        If corr == False, lat and lat are assumed to be edge coordinates (plotted with a small offset). The plotted image is (m-1) x (n-1).
        crange is a tuple giving lower and upper colormap limits.
        bgcolor is a str giving background color inside the polar region.
        """

        x,y =self._latlt2xy(lat, lt)
        if corr:
            xe = np.full(np.subtract(x.shape,(1,1)),np.nan)
            ye = np.full(np.subtract(y.shape,(1,1)),np.nan)
            for i in range(xe.shape[0]):
                for j in range(xe.shape[1]):
                    xe[i,j]=np.mean(x[i:i+2,j:j+2])
                    ye[i,j]=np.mean(y[i:i+2,j:j+2])
            x=xe
            y=ye
            data = image[1:-1,1 :-1]
        else:
            data = image[1:, :-1]

        ll = (x[1:,  :-1].flatten(), y[1:,  :-1].flatten())
        lr = (x[1:,   1:].flatten(), y[1:,   1:].flatten())
        ul = (x[:-1, :-1].flatten(), y[:-1, :-1].flatten())
        ur = (x[:-1,  1:].flatten(), y[:-1,  1:].flatten())

        vertsx = np.vstack((ll[0], lr[0], ur[0], ul[0])).T
        vertsy = np.vstack((ll[1], lr[1], ur[1], ul[1])).T

        iii = np.where(vertsx**2 + vertsy**2 <=1, True, False)
        iii = np.any(iii, axis = 1).nonzero()[0]

        vertsx = vertsx[iii]
        vertsy = vertsy[iii]

        verts = np.dstack((vertsx, vertsy))
        if 'cmap' in kwargs.keys():
            cmap = kwargs.pop('cmap')
        else:
            cmap = plt.cm.viridis
        data = np.ma.array(data, mask=np.isnan(data))
        coll = PolyCollection(verts, array=data.flatten()[iii], cmap = cmap, edgecolors='none', **kwargs)
        if crange is not None:
            coll.set_clim(crange[0], crange[1])

        if bgcolor is not None:
            radius = 2*((90 - self.minlat)/ (90 - self.minlat))
            bg = Ellipse([0, 0], radius, radius, zorder = 0, facecolor = bgcolor)
            self.ax.add_artist(bg)

        return self.ax.add_collection(coll)


    def ampereplot(self, keys, data, contour = False, crange = None, nlevels = 100, cbar = False, **kwargs):
        """ plot ampere data. keys is the columns of the ampere dataframes.
            data is an array of equal size, that will be plotted in grids of 1 degree by 1 hour size

            set cbar True to draw a color bar

            2021-7-1: Kalle: Added this function from an old version of polarsublot. It worked in 2016, but I have
                             no clue if it still works...
        """
        data = np.array(data)

        # MAKE VERTICES
        colats, lts = np.array(list(keys)).T
        lats = np.abs(90. - colats)

        ltres, latres = 1, 1

        if 'cmap' in kwargs.keys():
            cmap = kwargs['cmap']
            kwargs.pop('cmap')
        else:
            cmap = plt.cm.RdBu_r

        if not contour: # plot so that the grid cells are seen
            latl, ltl = lats, lts
            latu = latl + latres
            ltr  = ltl + ltres
            verts = np.empty((len(keys), 20, 2))

            for i in range(len(lts)):
                # convert corners to cartesian (lat 90 at origin, lat 0 at distance 1 from origin)
                la = np.linspace((ltl[i]-6)/12.*np.pi, (ltr[i]-6)/12.*np.pi, 10)
                ua = np.linspace((ltl[i]-6)/12.*np.pi, (ltr[i]-6)/12.*np.pi, 10)[::-1]
                verts[i, :10, :]  = (90 - latl[i])/(90 - self.minlat) * np.array([np.cos(la), np.sin(la)]).T
                verts[i,  10:, :] = (90 - latu[i])/(90 - self.minlat) * np.array([np.cos(ua), np.sin(ua)]).T

            # the sign of the data is changed to make the color scale consistent with AMPERE summary plots
            coll = PolyCollection(verts[np.isfinite(data)], array = data[np.isfinite(data)], cmap = cmap, edgecolors='none', **kwargs)
            if crange is not None:
                coll.set_clim(crange[0], crange[1])
            self.ax.add_collection(coll)

        else: # plot a smooth contour
            coll = self.contourf(lats, lts, data, cmap = cmap, levels = np.linspace(crange[0], crange[1], nlevels), extend = 'both', **kwargs)

        if cbar:

            clim = coll.get_clim()
            axpos = self.ax.get_position().bounds # position of current plot
            cbax = self.ax.figure.add_subplot(111)
            cbax.set_position([axpos[0] + axpos[2], axpos[1], 0.02, axpos[3]])
            #cbax.get_xaxis().set_visible(False)
            #cbax.get_yaxis().set_visible(False)
            RESOLUTION = 100
            colors = np.linspace(clim[0], clim[1], RESOLUTION + 1)
            colors -= colors.min()
            cbax.set_ylim(colors.min(), colors.max())
            verts = np.empty((RESOLUTION, 4, 2))
            for i in range(RESOLUTION):
                verts[i, :, :] = np.array([[0, 0, 1, 1], [colors[i], colors[i+1], colors[i+1], colors[i]]]).T

            cbarcoll = PolyCollection(verts, array = np.linspace(clim[0], clim[1], RESOLUTION), cmap = cmap, edgecolors='none')
            cbarcoll.set_clim(clim)
            cbax.add_collection(cbarcoll)

            cbax.set_xlabel(r'$\mu$A/m$^2$', labelpad = 1)
            cbax.set_yticks([colors.min(), 0 - clim[0], colors.max()])
            cbax.set_yticklabels([str(clim[0]), '0', str(clim[1])])
            cbax.set_xlim(0, 1)
            cbax.set_xticks([0, 1])
            cbax.set_xticklabels(['', ''])

        return coll


    def coastlines(self, time = None, mag = None, north = True, resolution = '50m', **kwargs):
        """ plot coastlines

        Coastline data are read from numpy savez files. These files are made
        with the download_coastlines.py script in the helper_scripts folder,
        which uses the cartopy module.

        Parameters
        ----------
        time: datetime, optional
            give a datetime to replace longitude with local time when
            plotting coastlines
        mag: apexpy.Apex, optional
            give an apexpy.Apex object to convert coastlines to magnetic
            apex coordinates. If None (default), coastlines will be
            plotted in geographic coordinates
        north: bool, optional
            set to False if you want coastlines plotted for the southern
            hemisphere. Default is True
        resolution: string, optional
            Set to '50m' or '110m' to adjust the resolution of the coastlines.
            These options correspond to options given to cartopy. Default is
            '50m', which is the highest resolution.
        **kwargs: dict, optional
            keywords passed to matplotlib's plot function
        """

        coastlines = np.load(datapath + 'coastlines_' + resolution + '.npz')

        segments = []
        for cl in coastlines:
            lat, lon = coastlines[cl]

            # convert lat and lon to magnetic if mag is given:
            if mag is not None:
                lat, lon = mag.geo2apex(lat, lon, mag.refh)

            if time is not None: # calculate local time if time is given
                if mag is None: # calculate geographic local time
                    sslat, sslon = subsol(time)
                    londiff = (lon - sslon + 180) % 360 - 180
                    lon = (180. + londiff)/15.
                else: # magnetic local time:
                    lon = mag.mlon2mlt(lon, time)
            else: # keep longitude - just divide by 15 to get unit hours
                lon = lon / 15

            if not north:
                lat = -1 * lat
            if not np.any(lat > self.minlat):
                continue

            iii = lat > self.minlat
            lat[~iii] = np.nan
            lon[~iii] = np.nan

            x, y = self._latlt2xy(lat, lon)

            segments.append(np.vstack((x, y)).T)

        collection = LineCollection(segments, **kwargs)
        return self.ax.add_collection(collection)



    def _latlt2xy(self, lat, lt, ignore_plot_limits=False):
        """

        Parameters
        ----------
        lt : TYPE
            DESCRIPTION.
        lat : TYPE
            DESCRIPTION.
        ignore_plot_limits : boolean, optional
            When True the conversion will ignore the limits of the plot allowing plotting outside
            the limits of the subplot, for example this is used in writeLTlabels. The default is False.

        Raises
        ------
        ValueError
            DESCRIPTION.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """

        np.seterr(invalid='ignore', divide='ignore')
        lt = np.array(lt) % 24
        lat = np.abs(np.array(lat))

        if lt.size!= lat.size:
            raise ValueError('x and y must be the same size')

        r = (90. - np.abs(lat))/(90. - self.minlat)
        a = (lt - 6.)/12.*np.pi

        x, y = r*np.cos(a), r*np.sin(a)

        if ignore_plot_limits:
            return x, y
        else:
            mask = np.ones_like(x)
            mask[(~eval(self.ltlims)) | (lat < self.minlat)] = np.nan
            return x * mask, y * mask

    def _xy2latlt(self, x, y, ignore_plot_limits=False):
        """
        convert x, y to lt, lat, where x**2 + y**2 = 1 corresponds to self.minlat

        Parameters
        ----------
        x : float/integer/list/array
            DESCRIPTION.
        y : TYPE
            DESCRIPTION.
        ignore_plot_limits : boolean, optional
            When True the conversion will ignore the limits of the plot allowing plotting outside
            the limits of the subplot, for example this is used in writeLTlabels. The default is False.

        Returns
        -------
        TYPE
            DESCRIPTION.
        TYPE
            DESCRIPTION.

        """
        x, y = np.array(x, ndmin = 1, dtype='float64'), np.array(y, ndmin = 1, dtype='float64') # convert to array to allow item assignment

        lat = 90 - np.sqrt(x**2 + y**2)*(90. - self.minlat)
        lt = np.arctan2(y, x)*12/np.pi + 6
        lt = lt % 24

        if ignore_plot_limits:
            return lat, lt
        else:
            mask = np.ones_like(lt)
            mask[(~eval(self.ltlims)) | (lat < self.minlat)] = np.nan
            return lat * mask, lt * mask


    def _northEastToCartesian(self, north, east, lt):
        a = (lt - 6)/12*np.pi # convert LT to angle with x axis (pointing from pole towards dawn)

        x1 = np.array([-north*np.cos(a), -north*np.sin(a)]) # arrow direction towards origin (northward)
        x2 = np.array([-east*np.sin(a),  east*np.cos(a)])   # arrow direction eastward

        return x1 + x2


    def make_format(current, lt_label='lt', lat_label='lat'):
    # current and other are axes
        def format_coord(x, y):
            # x, y are data coordinates
            # convert to display coords
            display_coord = current._xy2latlt(x, y)[::-1]
            string_original= 'x={:.2f}, y={:.2f}'.format(x, y)
            string_magnetic= f'{lt_label}={float(display_coord[0]):.2f}, {lat_label}={float(display_coord[1]):.2f}'
            
            
            return (string_original.ljust(20)+string_magnetic)
        return format_coord





# Helper functions
# ----------------
def date_to_doy(month, day, leapyear = False):
    """ return day of year (DOY) at given month, day
    """

    # check that month in [1, 12]
    if month < 1 or month > 12:
        raise ValueError('month not in [1, 12]')

    # check if day < 1
    if day < 1:
        raise ValueError('date2doy: day must not be less than 1')

    # check if day exceeds days in months
    days_in_month = np.array([0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31])

    if day > days_in_month[month]:
        if not (leapyear & (month == 2) & (day == 29)):
            raise ValueError('date2doy: day must not exceed number of days in month')

    cumdaysmonth = np.cumsum(days_in_month[:-1])

    # day of year minus possibly leap day:
    doy = cumdaysmonth[month - 1] + day

    # add 1 if leapyear and month >= 3:
    return doy + leapyear if month >= 3 else doy


def is_leapyear(year):
    """ Check for leapyear

    """
    if year % 400 == 0:
        return True

    if year % 100 == 0:
        return False

    if year % 4 == 0:
        return True

    else:
        return False


def subsol(time):
    """
    calculate subsolar point at given datetime(s)

    returns:
      subsol_lat  -- latitude of the subsolar point
      subsol_lon  -- longiutde of the subsolar point

    After Fortran code by: 961026 A. D. Richmond, NCAR

    Documentation from original code:
    Find subsolar geographic latitude and longitude from date and time.
    Based on formulas in Astronomical Almanac for the year 1996, p. C24.
    (U.S. Government Printing Office, 1994).
    Usable for years 1601-2100, inclusive.  According to the Almanac,
    results are good to at least 0.01 degree latitude and 0.025 degree
    longitude between years 1950 and 2050.  Accuracy for other years
    has not been tested.  Every day is assumed to have exactly
    86400 seconds; thus leap seconds that sometimes occur on December
    31 are ignored:  their effect is below the accuracy threshold of
    the algorithm.

    """

    year = time.year
    # day of year:
    doy  = date_to_doy(time.month, time.day, is_leapyear(year))
    # seconds since start of day:
    ut   = time.hour * 60.**2 + time.minute*60. + time.second

    yr = year - 2000

    if year >= 2100 or year <= 1600:
        raise ValueError('subsol.py: subsol invalid after 2100 and before 1600')

    nleap = np.floor((year-1601)/4.)
    nleap = nleap - 99

    # exception for years <= 1900:
    ncent = np.floor((year-1601)/100.)
    ncent = 3 - ncent
    if year <= 1900: nleap = nleap + ncent

    l0 = -79.549 + (-.238699*(yr-4*nleap) + 3.08514e-2*nleap)

    g0 = -2.472 + (-.2558905*(yr-4*nleap) - 3.79617e-2*nleap)

    # Days (including fraction) since 12 UT on January 1 of IYR:
    df = (ut/86400. - 1.5) + doy

    # Addition to Mean longitude of Sun since January 1 of IYR:
    lf = .9856474*df

    # Addition to Mean anomaly since January 1 of IYR:
    gf = .9856003*df

    # Mean longitude of Sun:
    l = l0 + lf

    # Mean anomaly:
    g = g0 + gf
    grad = g*np.pi/180.

    # Ecliptic longitude:
    lmbda = l + 1.915*np.sin(grad) + .020*np.sin(2.*grad)
    lmrad = lmbda*np.pi/180.
    sinlm = np.sin(lmrad)

    # Days (including fraction) since 12 UT on January 1 of 2000:
    n = df + 365.*yr + nleap

    # Obliquity of ecliptic:
    epsilon = 23.439 - 4.e-7*n
    epsrad  = epsilon*np.pi/180.

    # Right ascension:
    alpha = np.arctan2(np.cos(epsrad)*sinlm, np.cos(lmrad)) * 180./np.pi

    # Declination:
    delta = np.arcsin(np.sin(epsrad)*sinlm) * 180./np.pi

    # Subsolar latitude:
    sbsllat = delta

    # Equation of time (degrees):
    etdeg = l - alpha
    nrot = np.round(etdeg/360.)
    etdeg = etdeg - 360.*nrot

    # Apparent time (degrees):
    aptime = ut/240. + etdeg    # Earth rotates one degree every 240 s.

    # Subsolar longitude:
    sbsllon = 180. - aptime
    nrot = np.round(sbsllon/360.)
    sbsllon = sbsllon - 360.*nrot

    return sbsllat, sbsllon


def equal_area_grid(dr = 2, K = 0, M0 = 8, N = 20):
    """ function for calculating an equal area grid in polar coordinates

    Parameters
    ----------
    dr : float (optional)
        The latitudinal resolution of the grid. Default 2 degrees
    K : int (optional)
        This number determines the colatitude of the inner radius of the
        post poleward ring of grid cells (the pole is not inlcluded!). 
        It relates to this colatitude (r0) as r0/dr = (2K + 1)/2 => K = (2r0/dr - 1)/2.
        Default value is 0
    M0 : int (optional)
        The number of sectors in the most poleward ring. Default is 8
    N : int (optional)
        The number of rings to be included. This determiend how far 
        equatorward the grid extends. Typically if dr is changed from 2 to 1, 
        N should be doubled to reach the same latitude. Default is 20, which means
        that the equatorward edge of the grid is 89 - 20*2 = 49 degrees
        (the most poleward latitude is 89 with default values)

    Returns
    -------
    mlat : array
        Array of latitudes for the equatorward west ("lower left") corners of the grid
        cells.
    mlt : array
        Array of magnetic local times for the equatorward west ("lower left") corner
        of the grid cells.
    mltres : array
        width, in magnetic local time, of the cells with lower left corners described 
        by mlat and mlt. Notice that this width changes with latitude, while the 
        latitudinal width is fixed, determined by the `dr` parameter

    
    """

    r0 = dr * (2*K + 1)/2.

    assert M0 % (K + 1) == 0 # this must be fulfilled

    grid = {}

    M = M0
    grid[90 - r0 - dr] = np.linspace(0, 24 - 24./M, M) # these are the lower limits in MLT

    for i in range(N - 1):

        M = int(M *  (1 + 1./(K + i + 1.))) # this is the partion for i + 1

        grid[90 - (r0 + (i + 1)*dr) - dr] = np.linspace(0, 24 - 24./M, M) # these are the lower limits in MLT

    mlats = []
    mlts = []
    mltres = []
    for key in sorted(grid.keys()):
        mltres_ = sorted(grid[key])[1] - sorted(grid[key])[0]
        for mlt in sorted(grid[key]):
            mlats.append(key)
            mlts.append(mlt)
            mltres.append(mltres_)

    return np.array(mlats), np.array(mlts), np.array(mltres)


def get_h2d_bin_areas(minlats, maxlats, minlons, maxlons, haversine=True,
                      rearth=6.370949e3,
                      altitude=0,
                      spherical_rectangle=True,
                      do_extra_width_calc=True):
    # +
    # NAME:                        GET_H2D_BIN_AREAS
    #
    # PURPOSE:                     And now you wanna know the area, too?!?!?!
    #                              Well, you can have it--in km^2
    # CATEGORY:
    #
    # MODIFICATION HISTORY:      2016/03/12 Barnebarn
    #
    # Notáte bien—se dan las áreas en kilometros cuadrados
    # -

    ## BINEDGE1=Binedge1, BINEDGE2=Binedge2

    # breakpoint()

    # rearth = 6.370949e3+altitude
    # print("rearth:", rearth)

    # print("Want more precision? Try https://math.stackexchange.com/questions/1205927/how-to-calculate-the-area-covered-by-any-spherical-rectangle")
    areas = latlong_pair_area(
        minlons, minlats, maxlons, maxlats, rearth=rearth, haversine=haversine,
        altitude=altitude,
        spherical_rectangle=spherical_rectangle,
        do_extra_width_calc=do_extra_width_calc)  # use radius at 100 km

    return areas


def latlong_pair_area(lon1, lat1,
                      lon2, lat2,
                      rearth=6.370949e3,
                      altitude=0,
                      haversine=True,
                      spherical_rectangle=True,
                      do_extra_width_calc=True):
    """
    Get area (in km^2) of a rectangle on a sphere defined by two lat/lon pairs
    """
    # 2016/03/12 Spence needs this
    # Example: Surf. Area of Utah in two squares
    # ##:Upper square:
    # lonLat1  = [-114.043579, 41.951320]
    # lonLat2  = [-111.055298, 41.021355]
    # upArea   = latlong_pair_area(lonLat1[0],lonLat1[1],lonLat2[0],lonLat2[1])
    #
    # ##:Lower square:
    # lonLat1  = [-114.043579, 41.021355]
    # lonLat2  = [-109.046173, 37.009133]
    # downArea = latlong_pair_area(lonLat1[0],lonLat1[1],lonLat2[0],lonLat2[1])
    #
    # uparea+downarea = 212565.86 km^2
    # actual area of Utah = 84899 sq. mi. * (1.60934 km/mi)^2 = 219886.3 sq. km.
    # Error = 3.33% (7320 sq. km.). Not bad!

    # rearth += altitude

    height = geo_dist(lon1, lat1,
                      lon1, lat2,
                      rearth=rearth,
                      altitude=altitude,
                      haversine=haversine)

    if do_extra_width_calc:
        width1 = geo_dist(lon1, lat1,
                          lon2, lat1,
                          rearth=rearth,
                          altitude=altitude,
                          haversine=haversine)
        width2 = geo_dist(lon1, lat2,
                          lon2, lat2,
                          rearth=rearth,
                          altitude=altitude,
                          haversine=haversine)

        # width = np.max([width1, width2])

        # width = np.mean([width1, width2])
        width = np.mean(np.vstack([width1,width2]),axis=0)
    else:
        width = geo_dist(lon1, lat1,
                         lon2, lat1,
                         rearth=rearth,
                         altitude=altitude,
                         haversine=haversine)

    if spherical_rectangle:

        return 4*rearth**2.*np.arcsin(np.tan(width/rearth/2.)*np.tan(height/rearth/2.))
    else:
        # print("Vanlig")
        return height*width


def geo_dist(lon1, lat1,
             lon2, lat2,
             deg=True,
             rearth=6.370949e3,
             altitude=0,
             haversine=True):  # default to earth radius in km
    # MODIFICATION HISTORY:
    # 	Written by:	Daithi A. Stone (stoned@uvic.ca), 2000-06-29.
    #	Modified:	DAS, 2000-07-06 (removed LENGTH.pro, added
    #			DIMENSION.pro).
    #	Modified:	DAS, 2000-07-24 (added Degrad constant).
    #	Modified:	DAS, 2000-08-17 (coverted from Cosine to Haversine
    #			Haversine formula, added SINGLE keyword).
    #	Modified:	DAS, 2002-08-12 (complies with CONSTANTS.pro revision)
    #   Modified:       SMH, 2016-03-12 Made my own stuff
    #   Modified:       SMH, 2019-11-18 From IDL code

    rearth += altitude

    if deg:
        lat1, lat2, lon1, lon2 = map(np.deg2rad, (lat1, lat2, lon1, lon2))

    # Difference coordinates
    dlon = lon2 - lon1
    dlat = lat2 - lat1

    # breakpoint()

    if haversine:

        # ***********************************************************************
        # Haversine Formula

        # Main calculation
        a = (np.sin(dlat / 2.))**2 + np.cos(lat1) * \
            np.cos(lat2) * (np.sin(dlon / 2.))**2
        a = np.sqrt(a)

        # A fix if a>1
        id = np.where(a > 1)[0]

        if id.size > 0:
            a[id] = 1

        # Convert to distance
        dist = rearth * 2 * np.arcsin(a)

    else:
        # ***********************************************************************
        # Cosine Formula (not used, but I thought I would leave it in)

        # Convert to Cartesian coordinates
        x1 = np.cos(lon1) * np.cos(lat1)
        y1 = np.sin(lon1) * np.cos(lat1)
        z1 = np.sin(lat1)
        x2 = np.cos(lon2) * np.cos(lat2)
        y2 = np.sin(lon2) * np.cos(lat2)
        z2 = np.sin(lat2)

        # Direction cosine
        dx = x1*x2+y1*y2+z1*z2

        # A fix if the |cosine| > 1
        id = np.where(dx > 1)[0]
        if id.size > 0:
            dx[id] = 1.

        #Output (distance)
        dist = rearth*np.arccos(dx)

    return dist



""" tools that are useful for spherical harmonic analysis

    SHkeys       -- class to contain n and m - the indices of the spherical harmonic terms
    nterms       -- function which calculates the number of terms in a 
                    real expansion of a poloidal (internal + external) and toroidal expansion 
    legendre -- calculate associated legendre functions - with option for Schmidt semi-normalization



MIT License

Copyright (c) 2017 Karl M. Laundal

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import absolute_import, division
import numpy as np
import apexpy
from .mlt_utils import mlon_to_mlt
from builtins import range

import ppigrf

d2r = np.pi/180


DEFAULT = object()
refre = 6371.2 # reference radius

class SHkeys(object):
    """ container for n and m in spherical harmonics

        keys = SHkeys(Nmax, Mmax)

        keys will behave as a tuple of tuples, more or less
        keys['n'] will return a list of the n's
        keys['m'] will return a list of the m's
        keys[3] will return the fourth n,m tuple

        keys is also iterable

    """

    def __init__(self, Nmax, Mmax):
        keys = []
        for n in range(Nmax + 1):
            for m in range(Mmax + 1):
                keys.append((n, m))

        self.keys = tuple(keys)
        self.make_arrays()

    def __getitem__(self, index):
        if index == 'n':
            return [key[0] for key in self.keys]
        if index == 'm':
            return [key[1] for key in self.keys]

        return self.keys[index]

    def __iter__(self):
        for key in self.keys:
            yield key

    def __len__(self):
        return len(self.keys)

    def __repr__(self):
        return ''.join(['n, m\n'] + [str(key)[1:-1] + '\n' for key in self.keys])[:-1]

    def __str__(self):
        return ''.join(['n, m\n'] + [str(key)[1:-1] + '\n' for key in self.keys])[:-1]

    def setNmin(self, nmin):
        """ set minimum n """
        self.keys = tuple([key for key in self.keys if key[0] >= nmin])
        self.make_arrays()
        return self

    def MleN(self):
        """ set m <= n """
        self.keys = tuple([key for key in self.keys if abs(key[1]) <= key[0]])
        self.make_arrays()
        return self

    def Mge(self, limit):
        """ set m >= limit  """
        self.keys = tuple([key for key in self.keys if abs(key[1]) >= limit])
        self.make_arrays()
        return self

    def NminusModd(self):
        """ remove keys if n - m is even """
        self.keys = tuple([key for key in self.keys if (key[0] - abs(key[1])) % 2 == 1])
        self.make_arrays()
        return self

    def NminusMeven(self):
        """ remove keys if n - m is odd """
        self.keys = tuple([key for key in self.keys if (key[0] - abs(key[1])) % 2 == 0])
        self.make_arrays()
        return self

    def negative_m(self):
        """ add negative m to the keys """
        keys = []
        for key in self.keys:
            keys.append(key)
            if key[1] != 0:
                keys.append((key[0], -key[1]))
        
        self.keys = tuple(keys)
        self.make_arrays()
        
        return self


    def Shaveoff_last_k_nterms_for_m_gt(self,k,mmin):
        """For a given value of m (with m > mmin), remove the last k n terms.
        
        For example, if nmax = 65 and mmax = 3, 
        SHkeys(65, 3).setNmin(Nmin).MleN().Mge(0).Shaveoff_last_k_nterms_for_m_gt(k=2,mmin=0)
        will remove the n=64 and and n=65 keys for m > mmin = 0.
"""
        nmax = self.n.max()
        keepkey = lambda key: (key[0] <= (nmax - k) ) or (key[1] <= mmin)

        self.keys = tuple([key for key in self.keys if keepkey(key)])
        self.make_arrays()
        return self


    def Shaveoff_first_k_nterms_for_m_gt(self,k,mmin=-1):
        """ similar to Shaveoff_last_k_nterms_for_m_gt, but removes first k n-terms instead of last k n-terms """
        nmax = self.n.max()
        get_nprime = lambda key: np.maximum(key[1],1)
        keepkey = lambda key: (key[0] >= (get_nprime(key) + k) ) or (key[1] <= mmin)

        self.keys = tuple([key for key in self.keys if keepkey(key)])
        self.make_arrays()
        return self


    def make_arrays(self):
        """ prepare arrays with shape ( 1, len(keys) )
            these are used when making G matrices
        """

        if len(self) > 0:
            self.m = np.array(self)[:, 1][np.newaxis, :]
            self.n = np.array(self)[:, 0][np.newaxis, :]
        else:
            self.m = np.array([])[np.newaxis, :]
            self.n = np.array([])[np.newaxis, :]



def nterms(NT = 0, MT = 0, NVi = 0, MVi = 0, NVe = 0, MVe = 0):
    """ return number of coefficients in an expansion in real spherical harmonics of
        toroidal magnetic potential truncated at NT, MT
        poloidal magnetic potential truncated at NVi, MVi for internal sources
        poloidal magnetic potential truncated at NVe, MVe for external sources
    """

    return len(SHkeys(NT , MT ).setNmin(1).MleN().Mge(0)) + \
           len(SHkeys(NT , MT ).setNmin(1).MleN().Mge(1)) + \
           len(SHkeys(NVe, MVe).setNmin(1).MleN().Mge(0)) + \
           len(SHkeys(NVe, MVe).setNmin(1).MleN().Mge(1)) + \
           len(SHkeys(NVi, MVi).setNmin(1).MleN().Mge(0)) + \
           len(SHkeys(NVi, MVi).setNmin(1).MleN().Mge(1))



def legendre(nmax, mmax, theta, schmidtnormalize = True, keys = None):
    """ Calculate associated Legendre function P and its derivative

        Algorithm from "Spacecraft Attitude Determination and Control" by James Richard Wertz


        Parameters
        ----------
        nmax : int
            highest spherical harmonic degree
        mmax : int
            hightest spherical harmonic order
        theta : array, float
            colatitude in degrees (shape is not preserved)
        schmidtnormalize : bool, optional
            True if Schmidth seminormalization is wanted, False otherwise. Default True
        keys : SHkeys, optional
            If this parameter is set, an array will be returned instead of a dict. 
            The array will be (N, 2M), where N is the number of elements in `theta`, and 
            M is the number of keys. The first M columns represents a matrix of P values, 
            and the last M columns represent values of dP/dtheta

        Returns
        -------
        P : dict
            dictionary of Legendre function evalulated at theta. Dictionary keys are spherical harmonic
            wave number tuples (n, m), and values will have shape (N, 1), where N is number of 
            elements in `theta`. 
        dP : dict
            dictionary of Legendre function derivatives evaluated at theta. Dictionary keys are spherical
            harmonic wave number tuples (n, m), and values will have shape (N, 1), where N is number of 
            elements in theta. 
        PdP : array (only if keys != None)
            if keys != None, PdP is returned instaed of P and dP. PdP is an (N, 2M) array, where N is 
            the number of elements in `theta`, and M is the number of keys. The first M columns represents 
            a matrix of P values, and the last M columns represent values of dP/dtheta

    """

    theta = theta.flatten()[:, np.newaxis]

    P = {}
    dP = {}
    sinth = np.sin(d2r*theta)
    costh = np.cos(d2r*theta)

    if schmidtnormalize:
        S = {}
        S[0, 0] = 1.

    # initialize the functions:
    for n in range(nmax +1):
        for m in range(nmax + 1):
            P[n, m] = np.zeros_like(theta, dtype = np.float64)
            dP[n, m] = np.zeros_like(theta, dtype = np.float64)

    P[0, 0] = np.ones_like(theta, dtype = np.float64)
    for n in range(1, nmax +1):
        for m in range(0, min([n + 1, mmax + 1])):
            # do the legendre functions and derivatives
            if n == m:
                P[n, n]  = sinth * P[n - 1, m - 1]
                dP[n, n] = sinth * dP[n - 1, m - 1] + costh * P[n - 1, n - 1]
            else:

                if n == 1:
                    Knm = 0.
                    P[n, m]  = costh * P[n -1, m]
                    dP[n, m] = costh * dP[n - 1, m] - sinth * P[n - 1, m]

                elif n > 1:
                    Knm = ((n - 1)**2 - m**2) / ((2*n - 1)*(2*n - 3))
                    P[n, m]  = costh * P[n -1, m] - Knm*P[n - 2, m]
                    dP[n, m] = costh * dP[n - 1, m] - sinth * P[n - 1, m] - Knm * dP[n - 2, m]

            if schmidtnormalize:
                # compute Schmidt normalization
                if m == 0:
                    S[n, 0] = S[n - 1, 0] * (2.*n - 1)/n
                else:
                    S[n, m] = S[n, m - 1] * np.sqrt((n - m + 1)*(int(m == 1) + 1.)/(n + m))


    if schmidtnormalize:
        # now apply Schmidt normalization
        for n in range(1, nmax + 1):
            for m in range(0, min([n + 1, mmax + 1])):
                P[n, m]  *= S[n, m]
                dP[n, m] *= S[n, m]


    if keys is None:
        return P, dP
    else:
        Pmat  = np.hstack(tuple(P[key] for key in keys))
        dPmat = np.hstack(tuple(dP[key] for key in keys)) 
    
        return np.hstack((Pmat, dPmat))


def get_legendre_arrays(nmax, mmax, theta, keys,
                        schmidtnormalize = True,
                        negative_m = False,
                        minlat = 0,
                        return_full_P_and_dP=False):
    """ Schmidt normalization is optional - can be skipped if applied to coefficients 

        theta is colat [degrees]

        algorithm from "Spacecraft Attitude Determination and Control" by James Richard Wertz
        (http://books.google.no/books?id=GtzzpUN8VEoC&lpg=PP1&pg=PA781#v=onepage)
        ***NOTE: The algorithm calculates P^m_n (μ) = P^m_n(cosθ) and dP^m_n/dθ, but we wish
                 to instead calculate dP^m_n/dλ = -dP^m_n/dθ. Hence the application of a 
                 negative sign to dP^m_n here.

        must be tested for large n - this could be unstable
        sum over m should be 1 for all thetas

        Same as get_legendre, but returns a N by 2M array, where N is the size of theta,
        and M is the number of keys. The first half the columns correspond to P[n,m], with
        n and m determined from keys - an shkeys.SHkeys object - and the second half is dP[n,m]

        theta must be a column vector (N, 1)
    """


    P = {}
    dP = {}
    sinth = np.sin(d2r*theta)
    costh = np.cos(d2r*theta)

    if schmidtnormalize:
        S = {}
        S[0, 0] = 1.

    # initialize the functions:
    for n in range(nmax +1):
        for m in range(nmax + 1):
            P[n, m] = np.zeros_like(theta, dtype = np.float64)
            dP[n, m] = np.zeros_like(theta, dtype = np.float64)

    P[0, 0] = np.ones_like(theta, dtype = np.float64)
    P[0, 0][np.abs(90 - theta) < minlat] = 0
    for n in range(1, nmax +1):
        for m in range(0, min([n + 1, mmax + 1])):
            # do the legendre polynomials and derivatives
            if n == m:
                P[n, n]  = sinth * P[n - 1, m - 1]
                dP[n, n] = sinth * dP[n - 1, m - 1] + costh * P[n - 1, n - 1]
            else:

                if n == 1:
                    Knm = 0.
                    P[n, m]  = costh * P[n -1, m]
                    dP[n, m] = costh * dP[n - 1, m] - sinth * P[n - 1, m]

                elif n > 1:
                    Knm = ((n - 1)**2 - m**2) / ((2*n - 1)*(2*n - 3))
                    P[n, m]  = costh * P[n -1, m] - Knm*P[n - 2, m]
                    dP[n, m] = costh * dP[n - 1, m] - sinth * P[n - 1, m] - Knm * dP[n - 2, m]

            if schmidtnormalize:
                # compute Schmidt normalization
                if m == 0:
                    S[n, 0] = S[n - 1, 0] * (2.*n - 1)/n
                else:
                    S[n, m] = S[n, m - 1] * np.sqrt((n - m + 1)*(int(m == 1) + 1.)/(n + m))


    if schmidtnormalize:
        # now apply Schmidt normalization
        for n in range(1, nmax + 1):
            for m in range(0, min([n + 1, mmax + 1])):
                P[n, m]  *= S[n, m]
                dP[n, m] *= S[n, m]

    if negative_m:
        for n  in range(1, nmax + 1):
            for m in range(0, min([n + 1, mmax + 1])):
                P[n, -m]  = -1.**(-m) * factorial(n-m)/factorial(n+m) *  P[n, m]
                dP[n, -m] = -1.**(-m) * factorial(n-m)/factorial(n+m) * dP[n, m]

    if return_full_P_and_dP:
        return P, dP

    Pmat  = np.hstack(tuple(P[key] for key in keys))
    dPmat = np.hstack(tuple(dP[key] for key in keys)) 

    return np.hstack((Pmat, dPmat))


def get_legendre_arrays__Amatrix(nmax, mmax, theta, keys, A,
                                 schmidtnormalize = True,
                                 negative_m = False,
                                 minlat = 0,
                                 zero_thetas = 90.-np.array([47.,-47.]),
                                 return_full_P_and_dP=False,
                                 multiply_dP_by_neg1=False):
    """ Schmidt normalization is optional - can be skipped if applied to coefficients 

        theta is colat [degrees]

        algorithm from "Spacecraft Attitude Determination and Control" by James Richard Wertz
        (http://books.google.no/books?id=GtzzpUN8VEoC&lpg=PP1&pg=PA781#v=onepage)
        ***NOTE: The algorithm calculates P^m_n (μ) = P^m_n(cosθ) and dP^m_n/dθ, but we wish
                 to instead calculate dP^m_n/dλ = -dP^m_n/dθ. Hence the application of a 
                 negative sign to dP^m_n here.

        must be tested for large n - this could be unstable
        sum over m should be 1 for all thetas

        Same as get_legendre, but returns a N by 2M array, where N is the size of theta,
        and M is the number of keys. The first half the columns correspond to P[n,m], with
        n and m determined from keys - an shkeys.SHkeys object - and the second half is dP[n,m]

        theta must be a column vector (N, 1)
    """

    P = {}
    dP = {}

    sinth = np.sin(d2r*theta)
    costh = np.cos(d2r*theta)

    if schmidtnormalize:
        S = {}
        S[0, 0] = 1.

    # initialize the functions:
    for n in range(nmax +1):
        for m in range(nmax + 1):
            P[n, m] = np.zeros_like(theta, dtype = np.float64)
            dP[n, m] = np.zeros_like(theta, dtype = np.float64)

    P[0, 0] = np.ones_like(theta, dtype = np.float64)
    P[0, 0][np.abs(90 - theta) < minlat] = 0
    for n in range(1, nmax +1):
        for m in range(0, min([n + 1, mmax + 1])):
            # do the legendre polynomials and derivatives
            if n == m:
                P[n, n]  = sinth * P[n - 1, m - 1]
                dP[n, n] = sinth * dP[n - 1, m - 1] + costh * P[n - 1, n - 1]
            else:

                if n == 1:
                    Knm = 0.
                    P[n, m]  = costh * P[n -1, m]
                    dP[n, m] = costh * dP[n - 1, m] - sinth * P[n - 1, m]

                elif n > 1:
                    Knm = ((n - 1)**2 - m**2) / ((2*n - 1)*(2*n - 3))
                    P[n, m]  = costh * P[n -1, m] - Knm*P[n - 2, m]
                    dP[n, m] = costh * dP[n - 1, m] - sinth * P[n - 1, m] - Knm * dP[n - 2, m]

            if schmidtnormalize:
                # compute Schmidt normalization
                if m == 0:
                    S[n, 0] = S[n - 1, 0] * (2.*n - 1)/n
                else:
                    S[n, m] = S[n, m - 1] * np.sqrt((n - m + 1)*(int(m == 1) + 1.)/(n + m))


    if schmidtnormalize:
        # now apply Schmidt normalization
        for n in range(1, nmax + 1):
            for m in range(0, min([n + 1, mmax + 1])):
                P[n, m]  *= S[n, m]
                dP[n, m] *= S[n, m]

    if negative_m:
        for n  in range(1, nmax + 1):
            for m in range(0, min([n + 1, mmax + 1])):
                P[n, -m]  = -1.**(-m) * factorial(n-m)/factorial(n+m) *  P[n, m]
                dP[n, -m] = -1.**(-m) * factorial(n-m)/factorial(n+m) * dP[n, m]

    # Make fmn
    # f = {}
    # df = {}
    # for m in range(0,mmax + 1):
    #     for n in range (1, nmax + 1):
    
    #         f[n, m] = Tcoeff[n, m] * P[nmax, m] - Qtilde[n, m] * P[nmax-1, m] + P[n, m]
    #         df[n, m] = Tcoeff[n, m] * dP[nmax, m] - Qtilde[n, m] * dP[nmax-1, m] + dP[n, m]

    if multiply_dP_by_neg1:
        dP = {key:(-1)*dP[key] for key in dP.keys()}

    if return_full_P_and_dP:
        return P, dP

    Pmat  = np.hstack(tuple(P[key] for key in keys))
    dPmat = np.hstack(tuple(dP[key] for key in keys))
    # return Pmat, dPmat, A
    return np.hstack((Pmat@A, dPmat@A))


def get_R_arrays(nmax, mmax, theta, keys = None,
                 zero_thetas = 90.-np.array([47.,-47.]),
                 schmidtnormalize = True,
                 return_Q = False):#,
                 # minlat = 0,
                 # return_full_P_and_dP=False):
    """ Schmidt normalization is optional - can be skipped if applied to coefficients 

        theta is colat [degrees]

        algorithm from "Spacecraft Attitude Determination and Control" by James Richard Wertz
        (http://books.google.no/books?id=GtzzpUN8VEoC&lpg=PP1&pg=PA781#v=onepage)
        ***NOTE: The algorithm calculates P^m_n (μ) = P^m_n(cosθ) and dP^m_n/dθ, NOT dP^m_n/dλ

        must be tested for large n - this could be unstable
        sum over m should be 1 for all thetas

        Same as get_legendre, but returns a N by 2M array, where N is the size of theta,
        and M is the number of keys. The first half the columns correspond to P[n,m], with
        n and m determined from keys - an shkeys.SHkeys object - and the second half is dP[n,m]

        theta must be a column vector (N, 1)
    """

    assert len(zero_thetas) == 2
    # assert np.all([key in zero_keys for key in keys]),"'zero_keys' must include all keys in 'keys'!"

    zero_keys = SHkeys(nmax, mmax).setNmin(1).MleN().Mge(0)

    zero_thetas = zero_thetas.reshape((2,1))
    zero_T = get_legendre_arrays(nmax, mmax, zero_thetas, zero_keys, return_full_P_and_dP=True)
    zero_T_P = {key:zero_T[0][key] for key in zero_keys}
    zero_T_dP = {key:zero_T[1][key] for key in zero_keys}

    theta = theta.flatten()[:, np.newaxis]

    P = {}
    dP = {}

    sinth = np.sin(d2r*theta)
    costh = np.cos(d2r*theta)

    if schmidtnormalize:
        S = {}
        S[0, 0] = 1.

    # initialize the functions:
    for n in range(nmax +1):
        for m in range(nmax + 1):
            P[n, m] = np.zeros_like(theta, dtype = np.float64)
            dP[n, m] = np.zeros_like(theta, dtype = np.float64)

    P[0, 0] = np.ones_like(theta, dtype = np.float64)
    # P[0, 0][np.abs(90 - theta) < minlat] = 0
    for n in range(1, nmax +1):
        for m in range(0, min([n + 1, mmax + 1])):
            # do the legendre polynomials and derivatives
            if n == m:
                P[n, n]  = sinth * P[n - 1, m - 1]
                dP[n, n] = sinth * dP[n - 1, m - 1] + costh * P[n - 1, n - 1]
            else:

                if n == 1:
                    Knm = 0.
                    P[n, m]  = costh * P[n -1, m]
                    dP[n, m] = costh * dP[n - 1, m] - sinth * P[n - 1, m]

                elif n > 1:
                    Knm = ((n - 1)**2 - m**2) / ((2*n - 1)*(2*n - 3))
                    P[n, m]  = costh * P[n -1, m] - Knm*P[n - 2, m]
                    dP[n, m] = costh * dP[n - 1, m] - sinth * P[n - 1, m] - Knm * dP[n - 2, m]

            if schmidtnormalize:
                # compute Schmidt normalization
                if m == 0:
                    S[n, 0] = S[n - 1, 0] * (2.*n - 1)/n
                else:
                    S[n, m] = S[n, m - 1] * np.sqrt((n - m + 1)*(int(m == 1) + 1.)/(n + m))


    if schmidtnormalize:
        # now apply Schmidt normalization
        for n in range(1, nmax + 1):
            for m in range(0, min([n + 1, mmax + 1])):
                P[n, m]  *= S[n, m]
                dP[n, m] *= S[n, m]

    iplus = 0
    iminus = 1

    # Make Q terms
    Pratio_mu_plus = {key:zero_T_P[key][iplus][0]/zero_T_P[1,0][iplus][0] for key in zero_keys}

    Q = {key:P[key]-Pratio_mu_plus[key]*P[1,0] for key in zero_keys}
    dQ = {key:dP[key]-Pratio_mu_plus[key]*dP[1,0] for key in zero_keys}

    # Make R terms
    Q11_mu_minus = zero_T_P[1,1][iminus][0]-Pratio_mu_plus[1,1]*zero_T_P[1,0][iminus][0]

    R = {key:Q[key]*(1-Q[1,1]/Q11_mu_minus) for key in zero_keys}
    dR = {key:dQ[key]*(1-Q[1,1]/Q11_mu_minus)-Q[key]*dQ[1,1]/Q11_mu_minus for key in zero_keys}

    if keys is None:
        if return_Q:
            return R, dR, Q, dQ
        else:
            return R, dR
    else:

        # Pmat  = np.hstack(tuple(P[key] for key in keys))
        # dPmat = np.hstack(tuple(dP[key] for key in keys)) 
        # Qmat  = np.hstack(tuple(Q[key] for key in keys))
        # dQmat = np.hstack(tuple(dQ[key] for key in keys)) 
        Rmat  = np.hstack(tuple(R[key] for key in keys))
        dRmat = np.hstack(tuple(dR[key] for key in keys)) 
    
        if return_Q:
            Qmat  = np.hstack(tuple(Q[key] for key in keys))
            dQmat = np.hstack(tuple(dQ[key] for key in keys)) 
            
            return np.hstack((Rmat, dRmat)), np.hstack((Qmat, dQmat))

        else:
            return np.hstack((Rmat, dRmat))


def get_R_arrays__symm(nmax, mmax, theta, keys = None,
                 zero_thetas = 90.-np.array([47.,-47.]),
                 schmidtnormalize = True,
                 return_Q = False):#,
                 # minlat = 0,
                 # return_full_P_and_dP=False):
    """ get R arrays, but here we make sure that they are either symmetric or antisymmetric!
    """

    assert not return_Q,"Not implemented"

    zeros_0 = zero_thetas
    zeros_1 = 90.+(90.-zero_thetas)

    out0 = get_R_arrays(nmax, mmax, theta, keys = keys,
                        zero_thetas = zeros_0,
                        schmidtnormalize = schmidtnormalize,
                        return_Q = return_Q)

    out1 = get_R_arrays(nmax, mmax, theta, keys = keys,
                        zero_thetas = zeros_1,
                        schmidtnormalize = schmidtnormalize,
                        return_Q = return_Q)
    
    if keys is None:
        R0, dR0 = out0
        R1, dR1 = out1
        R = {key:(R0[key]+R1[key])/2 for key in R1.keys()}
        dR = {key:(dR0[key]+dR1[key])/2 for key in R1.keys()}
        
        return R, dR

    else:
        return (out0+out1)/2


def get_A_matrix__Ephizero(nmax, mmax,
                           zero_thetas = 90.-np.array([47.,-47.]),
                           return_all = False,
):
    
    assert len(zero_thetas) == 2

    zero_keys = SHkeys(nmax, mmax).setNmin(1).MleN().Mge(0)

    zero_thetas = zero_thetas.reshape((2,1))
    zero_T = get_legendre_arrays(nmax, mmax, zero_thetas, zero_keys, return_full_P_and_dP=True)
    zero_T_P = {key:zero_T[0][key] for key in zero_keys}
    zero_T_dP = {key:zero_T[1][key] for key in zero_keys}

    iplus = 0
    iminus = 1

    #Make Ptilde coeffs
    Ptilde = {}
    for m in range(0,mmax + 1):

        for n in range (1, nmax + 1):

            if (m == 0) or (n < m):
                Ptilde[n, m] = 0.
            else:
                Ptilde[n, m] = zero_T_P[n, m][iplus] / zero_T_P[nmax, m][iplus]


    #Make zero-Q
    zero_T_Q = {}
    for m in range(0,mmax + 1):
        for n in range (1, nmax + 1):
            if (m == 0) or (n < m):
                zero_T_Q[n, m] = 0.
            else:
                zero_T_Q[n, m] = zero_T_P[n, m][iminus] - Ptilde[n, m] * zero_T_P[nmax, m][iminus]
    

    #Make Qtilde coeffs
    Qtilde = {}
    for m in range(0,mmax + 1):
        for n in range (1, nmax + 1):

            if (m == 0) or (n < m):
                Qtilde[n, m] = 0.
            else:
                Qtilde[n, m] = zero_T_Q[n, m] / zero_T_Q[nmax-1, m]
            
    
    #Make T coeffs
    Tcoeff = {}
    for m in range(0,mmax + 1):
        for n in range (1, nmax + 1):

            if (m == 0) or (n < m):
                Tcoeff[n, m] = 0.
            else:
                Tcoeff[n, m] = Ptilde[nmax-1, m] * Qtilde[n, m] - Ptilde[n, m]

    # Need-'ems  for A matrix
    keys = SHkeys(nmax, mmax).setNmin(1).MleN().Mge(0)
    Ncoeffs = len(keys)
    narr = keys.n[0]
    marr = keys.m[0]
    
    # Now make A matrix
    A = np.zeros((Ncoeffs,Ncoeffs-2*mmax))
    zerorow = np.zeros(Ncoeffs-2*mmax)
    count = 0
    fixcount = 0
    for n in range (1, nmax + 1):
        for m in range(0,np.minimum(mmax + 1,n+1)):
    
            if (n >= (nmax-1)) and (m > 0):
                fixcount += 1
    
                tmprow = zerorow.copy()
    
                # get indices of (n', m) coefficients, n' ≤ nmax - 2, that we 
                # write the (nmax, m) or (nmax-1, m) coefficient in terms of
                fix_columns = np.where((narr < (nmax-1)) & (marr == m))[0]  
    
    
                # get values of n and m
                tmpn = narr[fix_columns]
                tmpm = marr[fix_columns]
    
                if n == nmax-1:
                    # do the thing for g^m_N-1
                    # tmprow[fix_columns] = -np.array([coeffs['Qtilde'][n,m] for n,m in zip(tmpn,tmpm)]).flatten()
                    tmprow[fix_columns] = -np.array([Qtilde[n,m] for n,m in zip(tmpn,tmpm)]).flatten()
    
                elif n == nmax:
                    # do the thing for g^m_N
                    # tmprow[fix_columns] = np.array([coeffs['Tcoeff'][n,m] for n,m in zip(tmpn,tmpm)]).flatten()
                    tmprow[fix_columns] = np.array([Tcoeff[n,m] for n,m in zip(tmpn,tmpm)]).flatten()
    
            else:
                tmprow = zerorow.copy()
                tmprow[count-fixcount] = 1
    
            A[count,:] = tmprow
    
            count +=1   
    
    if return_all:
        return A, dict(Qtilde=Qtilde,
                       Ptilde=Ptilde,
                       zero_T_Q=zero_T_Q,
                       Tcoeff=Tcoeff)
    else:
        return A


def get_A_matrix__potzero(nmax, mmax,
                          zero_thetas = 90.-np.array([47.,-47.]),
                          return_all = False,
):
    
    assert len(zero_thetas) == 2

    zero_keys = SHkeys(nmax, mmax).setNmin(1).MleN().Mge(0)

    zero_thetas = zero_thetas.reshape((2,1))
    zero_T = get_legendre_arrays(nmax, mmax, zero_thetas, zero_keys, return_full_P_and_dP=True)
    zero_T_P = {key:zero_T[0][key] for key in zero_keys}
    zero_T_dP = {key:zero_T[1][key] for key in zero_keys}

    iplus = 0
    iminus = 1

    #Make Ptilde coeffs
    Ptilde = {}
    for m in range(0,mmax + 1):

        nprime = np.maximum(m,1)

        for n in range(1, nmax + 1):

            if n < (nprime+1):
                Ptilde[n, m] = 0.
            else:
                Ptilde[n, m] = zero_T_P[n, m][iplus] / zero_T_P[nprime, m][iplus]


    #Make zero-Q
    zero_T_Q = {}
    for m in range(0,mmax + 1):

        nprime = np.maximum(m,1)

        for n in range (1, nmax + 1):
            if n < (nprime+1):
                zero_T_Q[n, m] = 0.
            else:
                zero_T_Q[n, m] = zero_T_P[n, m][iminus] - Ptilde[n, m] * zero_T_P[nprime, m][iminus]
    

    #Make Qtilde coeffs
    Qtilde = {}
    for m in range(0,mmax + 1):

        nprime = np.maximum(m,1)

        for n in range (1, nmax + 1):
            if n < (nprime + 2):
                Qtilde[n, m] = 0.
            else:
                Qtilde[n, m] = zero_T_Q[n, m] / zero_T_Q[nprime+1, m]
            
    #Make T coeffs
    Tcoeff = {}
    for m in range(0,mmax + 1):

        nprime = np.maximum(m,1)

        for n in range (1, nmax + 1):

            if n < (nprime + 2):
                Tcoeff[n, m] = 0.
            else:
                Tcoeff[n, m] = Ptilde[nprime+1, m] * Qtilde[n, m] - Ptilde[n, m]

    # Need-'ems  for A matrix
    keys = SHkeys(nmax, mmax).setNmin(1).MleN().Mge(0).Shaveoff_first_k_nterms_for_m_gt(2)
    Ntotcoeffs = len(zero_keys)
    Ncoeffs = len(keys)
    nallarr = zero_keys.n[0]
    mallarr = zero_keys.m[0]
    narr = keys.n[0]
    marr = keys.m[0]
    
    # Now make A matrix
    # A = np.zeros((Ntotcoeffs,Ntotcoeffs-2*(mmax+1)))
    # zerorow = np.zeros(Ntotcoeffs-2*(mmax+1))
    A = np.zeros((Ntotcoeffs,Ncoeffs))
    zerorow = np.zeros(Ncoeffs)
    count = 0
    fixcount = 0
    for n in range (1, nmax + 1):
        for m in range(0,np.minimum(mmax,n)+1):
    
            nprime = np.maximum(1,m)

            # if (n >= (nmax-1)) and (m > 0):
            if (n <= (nprime+1)):

                fixcount += 1
    
                tmprow = zerorow.copy()
    
                # get indices of (n, m) coefficients, n ≥ n' + 2, that we 
                # write the (n', m) or (n'+1, m) coefficient in terms of
                # where n' = max(m,1)
                # fix_columns = np.where((nallarr > (nprime+1)) & (mallarr == m))[0]  
                fix_columns = np.where((narr > (nprime+1)) & (marr == m))[0]  
    
                # get values of n and m
                tmpn = narr[fix_columns]
                tmpm = marr[fix_columns]
    
                if n == nprime:
                    # do the thing for g^m_n'
                    # tmprow[fix_columns] = -np.array([coeffs['Qtilde'][n,m] for n,m in zip(tmpn,tmpm)]).flatten()
                    # tmprow[fix_columns] = -np.array([Qtilde[n,m] for n,m in zip(tmpn,tmpm)]).flatten()
                    tmprow[fix_columns] = np.array([Tcoeff[n,m] for n,m in zip(tmpn,tmpm)]).flatten()
    
                elif n == nprime+1:
                    # do the thing for g^m_n'+1
                    # tmprow[fix_columns] = np.array([coeffs['Tcoeff'][n,m] for n,m in zip(tmpn,tmpm)]).flatten()
                    tmprow[fix_columns] = -np.array([Qtilde[n,m] for n,m in zip(tmpn,tmpm)]).flatten()
    
            else:
                tmprow = zerorow.copy()
                tmprow[count-fixcount] = 1
    
            A[count,:] = tmprow
    
            count +=1   
    
    if return_all:
        return A, dict(Qtilde=Qtilde,
                       Ptilde=Ptilde,
                       zero_T_Q=zero_T_Q,
                       Tcoeff=Tcoeff)
    else:
        return A


def getG_vel(glat, glon, height, time, epoch = 2015., h_R = 110.,
             NT = 65, MT = 3, 
             RR=6371.2,
             makenoise=False,
             toroidal_minlat=0,
             zero_lats=np.array([47.,-47.]),
             coords='geo'):
    """ all input arrays should be dask arrays with shape (N, 1), and with the same chunksize. """

    glat   = np.asarray(glat).flatten()
    glon   = np.asarray(glon).flatten()
    height = np.asarray(height).flatten()

    # convert to magnetic coords and get base vectors
    a = apexpy.Apex(epoch, refh = h_R)
    qlat, qlon = a.geo2qd(  glat.flatten(), glon.flatten(), height.flatten())
    alat, alon = a.geo2apex(glat.flatten(), glon.flatten(), height.flatten())
    f1, f2, f3, g1, g2, g3, d1, d2, d3, e1, e2, e3 = a.basevectors_apex(alat, alon, height, coords  = 'apex')
    d1e = d1[0].reshape(-1, 1)
    d1n = d1[1].reshape(-1, 1)
    d1u = d1[2].reshape(-1, 1)
    d2e = d2[0].reshape(-1, 1)
    d2n = d2[1].reshape(-1, 1)
    d2u = d2[2].reshape(-1, 1)
    e1e = e1[0].reshape(-1, 1)
    e1n = e1[1].reshape(-1, 1)
    e1u = e1[2].reshape(-1, 1)
    e2e = e2[0].reshape(-1, 1)
    e2n = e2[1].reshape(-1, 1)
    e2u = e2[2].reshape(-1, 1)

    from datetime import datetime
    Be, Bn, Bu = ppigrf.igrf(
        glon,
        glat,
        height,
        datetime(int(epoch),
                 np.clip(int((epoch-int(epoch))*12),1,12),
                 1)
    )
    B0IGRF = np.sqrt(Be**2+Bn**2+Bu**2)

    D = np.sqrt( (d1n*d2u-d1u*d2n)**2 + \
                 (d1u*d2e-d1e*d2u)**2 + \
                 (d1e*d2n-d1n*d2e)**2)

    Be3_in_Tesla = (B0IGRF/D.flatten()/1e9).T

    # calculate magnetic local time
    phi = mlon_to_mlt(qlon, time, a.year)[:, np.newaxis]*15 # multiply by 15 to get degrees

    # turn the coordinate arrays into column vectors:
    alat, h = map(lambda x: x.flatten()[:, np.newaxis], [alat, height])

    # generate spherical harmonic keys    
    keys = {} # dictionary of spherical harmonic keys
    keys['cos_T'] = SHkeys(NT, MT).setNmin(1).MleN().Mge(0).Shaveoff_first_k_nterms_for_m_gt(2)
    keys['sin_T'] = SHkeys(NT, MT).setNmin(1).MleN().Mge(1).Shaveoff_first_k_nterms_for_m_gt(2)

    fullkeys = {} # dictionary of spherical harmonic keys
    fullkeys['cos_T'] = SHkeys(NT, MT).setNmin(1).MleN().Mge(0)
    fullkeys['sin_T'] = SHkeys(NT, MT).setNmin(1).MleN().Mge(1)

    m_cos_T = keys['cos_T'].m
    m_sin_T = keys['sin_T'].m

    # generate Legendre matrices - first get dicts of arrays, and then stack them in the appropriate fashion
    if makenoise: print( 'Calculating Legendre functions. alat shape and chunks:', alat.shape, alat.chunks)
    assert len(zero_lats) == 2
    A = get_A_matrix__potzero(NT, MT,
                              zero_thetas = 90.-zero_lats)

    P_T = get_legendre_arrays__Amatrix(NT, MT, 90 - alat, fullkeys['cos_T'], A,
                                       minlat = toroidal_minlat,
                                       zero_thetas = 90.-zero_lats,
                                       multiply_dP_by_neg1=True)


    P_cos_T  =  P_T[:, :len(keys['cos_T']) ] # split
    dP_cos_T = P_T[:,  len(keys['cos_T']):]

    if makenoise: print( 'P, dP cos_T size and chunks', P_cos_T.shape, dP_cos_T.shape)#, P_cos_T.chunks, dP_cos_T.chunks
    P_sin_T  =  P_cos_T[ :, keys['cos_T'].m.flatten() != 0] 
    dP_sin_T =  dP_cos_T[:, keys['cos_T'].m.flatten() != 0]

    if makenoise: print( 'P, dP sin_T size and chunks', P_sin_T.shape, dP_sin_T.shape, P_sin_T.chunks[0], dP_sin_T.chunks[1])

    # trig matrices:
    cos_T  =  np.cos(phi * d2r * m_cos_T)#.rechunk((phi.chunks[0], m_cos_T.shape[1]))
    sin_T  =  np.sin(phi * d2r * m_sin_T)#.rechunk((phi.chunks[0], m_sin_T.shape[1]))

    dcos_T = -np.sin(phi * d2r * m_cos_T)#.rechunk((phi.chunks[0], m_cos_T.shape[1]))
    dsin_T =  np.cos(phi * d2r * m_sin_T)#.rechunk((phi.chunks[0], m_sin_T.shape[1]))

    if makenoise: print( cos_T.shape, sin_T.shape)

    cos_alat   = np.cos(alat * d2r)

    sinI  = 2 * np.sin( alat * d2r )/np.sqrt(4 - 3*cos_alat**2)

    R = (RR + h_R)                   # DON'T convert from km to m; this way potential is in kV

    # matrices with partial derivatives in MA coordinates:
    dT_dalon  = np.hstack(( P_cos_T * dcos_T * m_cos_T,  P_sin_T * dsin_T * m_sin_T))
    dT_dalat  = np.hstack((dP_cos_T *  cos_T          , dP_sin_T *  sin_T          ))

    # Stuff for ve1 and ve2, cf Equation 19 in Hatch et al (2023)
    # Remember that because Ed1 = -1/[ (R_E+h_R)*cos_alat ] dPhi/dphi [cf Eq 16], the negative sign in Eq 19 disappears
    # Hence no negative signs in expression for G matrix below
    V_E1 = RR/(R * Be3_in_Tesla * 1000) * (1 / sinI     * dT_dalat)
    V_E2 = RR/(R * Be3_in_Tesla * 1000) * (1 / cos_alat * dT_dalon)

    if coords == 'apex':
        G = np.vstack((V_E1, V_E2))

        return G

    elif coords == 'geo':

        v_e = V_E1 * e1e + V_E2 * e2e
        v_n = V_E1 * e1n + V_E2 * e2n
        v_u = V_E1 * e1u + V_E2 * e2u
    
        # G matrix for geographic components of E-field, cf Equation 15 in Hatch et al (2023)
        G     = np.vstack((v_e, v_n, v_u))
    
        return G


def getG_E(glat, glon, height, time, epoch = 2015., h_R = 110.,
           NT = 65, MT = 3, 
           RR=6371.2,
           makenoise=False,
           toroidal_minlat=0,
           zero_lats=np.array([47.,-47.]),
           coords='geo'):
    """ all input arrays should be dask arrays with shape (N, 1), and with the same chunksize. """

    assert coords in ['geo', 'apex']

    glat   = np.asarray(glat).flatten()
    glon   = np.asarray(glon).flatten()
    height = np.asarray(height).flatten()

    # convert to magnetic coords and get base vectors
    a = apexpy.Apex(epoch, refh = h_R)
    qlat, qlon = a.geo2qd(  glat.flatten(), glon.flatten(), height.flatten())
    alat, alon = a.geo2apex(glat.flatten(), glon.flatten(), height.flatten())
    f1, f2, f3, g1, g2, g3, d1, d2, d3, e1, e2, e3 = a.basevectors_apex(alat, alon, height, coords  = 'apex')
    d1e = d1[0].reshape(-1, 1)
    d1n = d1[1].reshape(-1, 1)
    d1u = d1[2].reshape(-1, 1)
    d2e = d2[0].reshape(-1, 1)
    d2n = d2[1].reshape(-1, 1)
    d2u = d2[2].reshape(-1, 1)

    # calculate magnetic local time
    phi = mlon_to_mlt(qlon, time, a.year)[:, np.newaxis]*15 # multiply by 15 to get degrees

    # turn the coordinate arrays into column vectors:
    alat, h = map(lambda x: x.flatten()[:, np.newaxis], [alat, height])

    # generate spherical harmonic keys    
    keys = {} # dictionary of spherical harmonic keys
    keys['cos_T'] = SHkeys(NT, MT).setNmin(1).MleN().Mge(0).Shaveoff_first_k_nterms_for_m_gt(2)
    keys['sin_T'] = SHkeys(NT, MT).setNmin(1).MleN().Mge(1).Shaveoff_first_k_nterms_for_m_gt(2)

    fullkeys = {} # dictionary of spherical harmonic keys
    fullkeys['cos_T'] = SHkeys(NT, MT).setNmin(1).MleN().Mge(0)
    fullkeys['sin_T'] = SHkeys(NT, MT).setNmin(1).MleN().Mge(1)

    m_cos_T = keys['cos_T'].m
    m_sin_T = keys['sin_T'].m

    # generate Legendre matrices - first get dicts of arrays, and then stack them in the appropriate fashion
    if makenoise: print( 'Calculating Legendre functions. alat shape and chunks:', alat.shape, alat.chunks)
    assert len(zero_lats) == 2
    A = get_A_matrix__potzero(NT, MT,
                              zero_thetas = 90.-zero_lats)

    P_T = get_legendre_arrays__Amatrix(NT, MT, 90 - alat, fullkeys['cos_T'], A,
                                       minlat = toroidal_minlat,
                                       zero_thetas = 90.-zero_lats,
                                       multiply_dP_by_neg1=True)


    P_cos_T  =  P_T[:, :len(keys['cos_T']) ] # split
    dP_cos_T = P_T[:,  len(keys['cos_T']):]

    if makenoise: print( 'P, dP cos_T size and chunks', P_cos_T.shape, dP_cos_T.shape)#, P_cos_T.chunks, dP_cos_T.chunks
    P_sin_T  =  P_cos_T[ :, keys['cos_T'].m.flatten() != 0] 
    dP_sin_T =  dP_cos_T[:, keys['cos_T'].m.flatten() != 0]

    if makenoise: print( 'P, dP sin_T size and chunks', P_sin_T.shape, dP_sin_T.shape, P_sin_T.chunks[0], dP_sin_T.chunks[1])

    # trig matrices:
    cos_T  =  np.cos(phi * d2r * m_cos_T)#.rechunk((phi.chunks[0], m_cos_T.shape[1]))
    sin_T  =  np.sin(phi * d2r * m_sin_T)#.rechunk((phi.chunks[0], m_sin_T.shape[1]))

    dcos_T = -np.sin(phi * d2r * m_cos_T)#.rechunk((phi.chunks[0], m_cos_T.shape[1]))
    dsin_T =  np.cos(phi * d2r * m_sin_T)#.rechunk((phi.chunks[0], m_sin_T.shape[1]))

    if makenoise: print( cos_T.shape, sin_T.shape)

    cos_alat   = np.cos(alat * d2r)

    sinI  = 2 * np.sin( alat * d2r )/np.sqrt(4 - 3*cos_alat**2)

    R = (RR + h_R)                   # DON'T convert from km to m; this way potential is in kV

    # matrices with partial derivatives in MA coordinates:
    dT_dalon  = np.hstack(( P_cos_T * dcos_T * m_cos_T,  P_sin_T * dsin_T * m_sin_T))
    dT_dalat  = np.hstack((dP_cos_T *  cos_T          , dP_sin_T *  sin_T          ))

    E_D1 = - RR/R * (1 / cos_alat * dT_dalon)
    E_D2 =   RR/R * (1 / sinI     * dT_dalat)

    if coords == 'apex':

        G = np.vstack((E_D1, E_D2))

        return G

    elif coords == 'geo':

        E_e = E_D1 * d1e + E_D2 * d2e
        E_n = E_D1 * d1n + E_D2 * d2n
        E_u = E_D1 * d1u + E_D2 * d2u
    
        # G matrix for geographic components of E-field, cf Equation 15 in Hatch et al (2023)
        G     = np.vstack((E_e, E_n, E_u))
    
        return G
    



# IPython log file

import numpy as np

from pyswipe.swipe import SWIPE

plotdir = '/plotdir/'

import matplotlib as mpl
mplBkgrnd = 'QtAgg'
mpl.use(mplBkgrnd)

mpl.rcParams.update({'text.color': 'k'})
mpl.rcParams.update({'axes.labelcolor': 'k'})
mpl.rcParams.update({'xtick.color': 'k'})
mpl.rcParams.update({'ytick.color': 'k'})
mpl.rcParams.update({'font.size': 14})
# mpl.rcParams.update({'font.family': 'sans-serif'})
# mpl.rcParams.update({'font.sans-serif': 'Arial'})
mpl.rcParams.update({'text.usetex': False})

mpl.rcParams.update({'figure.figsize': [10.0, 8.0]})
mpl.rcParams.update({'savefig.directory': plotdir})

import matplotlib.pyplot as plt
plt.ion()

d2r = np.pi/180


# FINE diffs
dptilts = [-25,0,25]

# BTs = [3,5]                          # nT
# BTs = [1,7]                          # nT
BTs = np.arange(0.,10.1,0.2)
cas = np.arange(-180,180,15)

# TEST
BTs = np.arange(1.,10.1,1.0)
cas = np.arange(-180,180,30)

# TEST
BTs = np.arange(1.,10.1,0.5)
cas = np.arange(-180,180,15)

# BTs = np.arange(1.,10.1,1.0)
# cas = np.arange(-180,180,30)

CA, BT = np.meshgrid(cas,BTs,indexing='ij')

print(f"N iters: {CA.size}")

swvel = 450
f107val = 120

savefig = True

minlat = 60

dr = 1                          # deltalat for vector grid, which is equal-area

##############################
#Grid area for integrated EM calc
# Ripped from pyamps.amps.get_integrated_upward_current
REFRE = 6371.2 # Reference radius used in geomagnetic modeling

swipen = SWIPE(-swvel,0.,0.,dptilts[0],f107val,
               minlat=minlat,
               dr=dr)

# get surface area element in each cell:
mlat, mlt = swipen.scalargrid
mlt_sorted = np.sort(np.unique(mlt))
mltres = (mlt_sorted[1] - mlt_sorted[0]) * np.pi / 12
mlat_sorted = np.sort(np.unique(mlat))
mlatres = (mlat_sorted[1] - mlat_sorted[0]) * np.pi / 180
R = (REFRE + swipen.height) * 1e3  # radius in meters
dS = R**2 * np.cos(mlat * np.pi/180) * mlatres * mltres

dS_n,dS_s = np.split(dS, 2)
assert np.array_equal(dS_n,dS_s),"Uh oh, calculations later assume dS_n == dS_s!"
dS = dS_n.ravel()

mlat_n, mlt_n = map(lambda x:np.split(x, 2)[0], swipen.scalargrid)
mlat_s, mlt_s = map(lambda x:np.split(x, 2)[1], swipen.scalargrid)


##############################
#Loop

# lists for holding CPCP for each bz value
dPhiNss, dPhiSss = [], []
locnss, locsss = [], []
for dptilt in dptilts:

    # lists for holding CPCP in each hemisphere and the locations of extrema
    dPhiNs, dPhiSs = [], []
    locns, locss = [], []
    for j,(_ca,B) in enumerate(zip(CA.ravel(),BT.ravel())):

        by, bz = np.sin(_ca * d2r) * B, np.cos(_ca * d2r) * B

        if j == 0:
            swipen = SWIPE(-swvel,by,bz,dptilt,f107val,
                           minlat=minlat,
                           dr=dr)

            swipes = SWIPE(-swvel,-by,bz,-dptilt,f107val,
                           minlat=minlat,
                           dr=dr)

        else:
            swipen.update_model(-swvel, by, bz, dptilt, f107val)
            swipes.update_model(-swvel, -by, bz, -dptilt, f107val)

        phin      = swipen.get_emwork(mlat_n, mlt_n).ravel()
        phis      = swipes.get_emwork(mlat_s, mlt_s).ravel()

        dPhiN = np.sum(phin*dS)/1e3/1e9  # 1e3 to junk 'milli' prefix in mW/m², 1e12 to go to GW
        dPhiS = np.sum(phis*dS)/1e3/1e9

        print(j,_ca,B,f"{dPhiN:6.1f}",f"{dPhiS:6.1f}",f"{dPhiN-dPhiS:6.1f}")

        dPhiNs.append(dPhiN)
        dPhiSs.append(dPhiS)

    dPhiNs = np.array(dPhiNs)
    dPhiSs = np.array(dPhiSs)

    dPhiNs = dPhiNs.reshape(CA.shape)
    dPhiSs = dPhiSs.reshape(CA.shape)

    dPhiNss.append(dPhiNs)
    dPhiSss.append(dPhiSs)

########################################
#NH/SH MASTER ASYMM COEFF 
cmap = 'magma'
cmap2 = 'bwr'
# labs = ['NH','SH',r"2*(NH-SH)/(NH+SH)"]
labs = [r"$W_N(B_y, B_z, \Psi)$",
        r"$W_S(-B_y, B_z, -\Psi)$",
        r'$A_{EM} = 2\frac{W_N - W_S}{W_N + W_S}$']
seaslabs = ['Local\nWinter','Equinox','Local\nSummer']
seaslabs2 = ['W','E','S']

extend = 'max'


CAuse = CA.copy()
BTuse = BT.copy()
xticks = [-180,-90,0,90,180]
xtickstr = ['-180','-90','0','90','']
shading = None

doroll = True

rollfunc = lambda x: np.roll(x,-int(cas.size//2),axis=0)
if doroll:

    CAmod = CA.copy()
    CAmod[CAmod < 0] = CAmod[CAmod < 0] + 360


    CAuse = CAmod

    BTuse = BT.copy()
    

    CAuse = rollfunc(CAuse)
    BTuse = rollfunc(BTuse)

    CAuse = np.vstack([CAuse,np.array([360]*BTs.size)])
    BTuse = np.vstack([BTuse,BTuse[0,:]])

    CAuse = np.concatenate([CAuse.T,CAuse[:,0][:,np.newaxis].T]).T
    BTuse = np.broadcast_to(np.append(BTs-np.diff(BTs)[0]/2,BTs[-1]+np.diff(BTs)[0]/2),CAuse.shape)
    shading = 'flat'
    xticks = [0,90,180,270,360]
    xtickstr = ['0','90','180','270','']

vmax = 0
vmax2 = 0
for j,dptilt in enumerate(dptilts):

    dPhiNs = dPhiNss[j]
    dPhiSs = dPhiSss[j]

    vmin = 0
    tvmax = np.max([dPhiNs,dPhiSs])
    
    vmax = np.max([tvmax,vmax])

    coeff = (dPhiNs-dPhiSs)/ ( (dPhiNs+dPhiSs) / 2 )

    tvmax2 = np.max(np.abs([np.max(coeff),np.min(coeff)]))
    vmax2 = np.max([tvmax2,vmax2])
    vmin2 = -vmax2


vmax = 150

fig = plt.figure(14,figsize = (11, 10))

plt.clf()

plt.subplots_adjust(top=0.95,bottom=0.08,left=0.13, wspace=0.3, hspace=0.4,right=0.95)
for j,dptilt in enumerate(dptilts):

    dPhiNs = dPhiNss[j].copy()
    dPhiSs = dPhiSss[j].copy()

    if doroll:
        dPhiNs = rollfunc(dPhiNs)
        dPhiSs = rollfunc(dPhiSs)

    coeff = (dPhiNs-dPhiSs)/ ( (dPhiNs+dPhiSs) / 2 )

    ax0 = plt.subplot2grid((4, 14), (j,  0), colspan = 4)
    ax1 = plt.subplot2grid((4, 14), (j,  4), colspan = 4)
    ax2 = plt.subplot2grid((4, 14), (j,  8), colspan = 4)

    plt.sca(ax0)
    # ax0.set_aspect('equal')
    if j == 0:
        ax0.set_title(labs[0])
    if j == 2:
        plt.xlabel(r"$\theta_c$ [deg]")
    plt.ylabel("$B_T$ [nT]")
    imn = plt.pcolormesh(CAuse,BTuse,dPhiNs,vmin=vmin,vmax=vmax,cmap=cmap,
                         shading=shading)
    
    plt.sca(ax1)
    # ax1.set_aspect('equal')
    if j == 0:
        ax1.set_title(labs[1])
    if j == 2:
        plt.xlabel(r"$\theta_c$ [deg]")
    _ = [ylab.set_visible(False) for ylab in ax1.yaxis.get_ticklabels()]
    ims = plt.pcolormesh(CAuse,BTuse,dPhiSs,vmin=vmin,vmax=vmax,cmap=cmap,
                         shading=shading)
    
    plt.sca(ax2)
    # ax2.set_aspect('equal')
    if j == 0:
        ax2.set_title(labs[2])
    if j == 2:
        plt.xlabel(r"$\theta_c$ [deg]")
    _ = [ylab.set_visible(False) for ylab in ax2.yaxis.get_ticklabels()]
    imd = plt.pcolormesh(CAuse,BTuse,coeff,vmin=vmin2,vmax=vmax2,cmap=cmap2,
                         shading=shading)
    
    # if j == 0:
    figtxtcoords = fig.transFigure.inverted().transform(ax0.transAxes.transform((0.,0.5)))
    figtxtcoords[0] -= 0.1
    fig.text(*figtxtcoords,seaslabs[j],rotation='vertical',va='center',ha='center',fontsize=18)

    for axl in [ax0,ax1,ax2]:
        axl.set_xticks(xticks,labels=xtickstr)

axc  = plt.subplot2grid((4, 100), (0, 88), rowspan=3, colspan = 3)
axc2 = plt.subplot2grid((4, 100), (0, 96), rowspan=3, colspan = 3)

cb1 = plt.colorbar(imn,cax=axc,extend=extend)
cb2 = plt.colorbar(imd,cax=axc2)
    

# Now master panel
coeffs = [(dPhiNs-dPhiSs)/ ( (dPhiNs+dPhiSs) / 2 ) for dPhiNs,dPhiSs in zip(dPhiNss,dPhiSss)]

avgs = [np.mean(coeff,axis=0) for coeff in coeffs]

avgsphic = [np.mean(coeff,axis=1) for coeff in coeffs]

ax = plt.subplot2grid((4, 14), (3, 0), colspan = 7)
ax2 = plt.subplot2grid((4, 14), (3, 7), colspan = 7)

linestyles = ['-','-','-']
colors = ['black','gray','lightgray']
linewidths = [2,3,4]

zorders = [10,5,0]

## Average over clock angle
plt.sca(ax)
plt.xlabel("$B_T$ [nT]")
plt.ylabel(r'$A_{EM}$')

_ = ax.spines['right'].set_visible(False)
_ = ax.spines['top'].set_visible(False)

# Only show ticks on the left and bottom spines
_ = ax.yaxis.set_ticks_position('left')
_ = ax.xaxis.set_ticks_position('bottom')

for j,dptilt in enumerate(dptilts):
    
    idx = j
    avg = avgs[idx]

    ax.plot(BTs,avg,label=seaslabs2[idx],color=colors[idx],
            linestyle=linestyles[idx],
            lw=linewidths[idx],
            zorder=zorders[idx])

ax.legend(ncols=3)


## Average over B_T
plt.sca(ax2)
plt.xlabel(r"$\theta_c$ [deg]")

_ = ax2.spines['right'].set_visible(False)
_ = ax2.spines['top'].set_visible(False)

# Only show ticks on the left and bottom spines
_ = ax2.yaxis.set_ticks_position('left')
_ = ax2.xaxis.set_ticks_position('bottom')

for j,dptilt in enumerate(dptilts):
    
    idx = j
    avg = avgsphic[idx]

    showie = rollfunc(np.append(avg,avg[0])) if doroll else np.append(avg,avg[0])

    ax2.plot(CAuse[:,0],showie,label=seaslabs[idx],color=colors[idx],linestyle=linestyles[idx],lw=linewidths[idx],
             zorder=zorders[idx])

ax.set_ylim(ax2.get_ylim())
tickleme = lambda i,x: "{:.1f}".format(x) if (((i+1) %2) == 0) else " "
ax.set_yticks(ax2.get_yticks(),labels=[tickleme(ix,x) for ix,x in enumerate(ax2.get_yticks())])

ax2.set_yticks(ax2.get_yticks(),labels=[" "]*len(ax2.get_yticks()))
ax2.set_xticks([0,90,180,270,360])


if savefig:
    tottitl = f"EMWork\nF10.7={f107val}"

    fname = "MASTER2_"+tottitl.replace('$','').replace('_','').replace(' ','').replace('$^\circ$','deg')+'.png'
    fname = fname.replace("\\Psi","ψ").replace("^\\circ","deg").replace("\n","_").replace("{SW}","sw").replace("[","_").replace("]","").replace('CPCP_','')
    print("Saving fig to "+fname)
    plt.savefig(plotdir+fname,dpi=300)
        
        
    
